"""
Part of the ProbLog distribution.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import glob
import os
import subprocess

# Load general Python modules
import sys
import tempfile
import traceback

import resource

# Load ProbLog modules
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
from problog.program import PrologString
from problog.ddnnf_formula import DDNNF
from problog.formula import LogicFormula

# Which compiled knowledge format to use? (SDD or DDNNF)
KNOWLEDGE = DDNNF

DEFAULT_PORT = 8000
DEFAULT_TIMEOUT = 60
DEFAULT_MEMOUT = 1.0  # gigabyte

# Load Python standard web-related modules (based on Python version)
import json
import mimetypes

if sys.version_info.major == 2:
    import BaseHTTPServer
    import urlparse

    def toBytes(string):
        return bytes(string)

    import urllib

    url_decode = urllib.unquote_plus
else:
    import http.server as BaseHTTPServer
    import urllib.parse as urlparse

    def toBytes(string):
        return bytes(string, "UTF-8")

    url_decode = urlparse.unquote_plus

# Contains special URL paths. Initialize with @handle_url
PATHS = {}


class handle_url(object):
    """Decorator for adding a handler for a URL.

    Example: to add a handler for GET /hello?name=World

    @handle_url('/hello')
    def hello(name) :
        return 200, 'text/plain', 'Hello %s!' % name
    """

    def __init__(self, path):
        self.path = path

    def __call__(self, f):
        PATHS[self.path] = f


def root_path(relative_path):
    """Translate URL path to file system path."""
    return os.path.abspath(
        os.path.join(os.path.dirname(__file__), relative_path.lstrip("/"))
    )


def model_path(*args):
    """Translate model path to file system path."""
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "../test/", *args))


def call_process(cmd, timeout, memout):
    """Call a subprocess with time and memory restrictions.

        Note:
            - the time restriction only works on Linux and Mac OS X
            - the memory restriction only works on Linux

        :param:timeout: CPU time in seconds
        :param:memout: Maximum memory in bytes
        :return: Output of the subprocess.

    """

    def setlimits():
        resource.setrlimit(resource.RLIMIT_CPU, (timeout, timeout))
        resource.setrlimit(resource.RLIMIT_AS, (memout, memout))

    return subprocess.check_output(cmd, preexec_fn=setlimits)


@handle_url("/problog")
def run_problog(model):
    """Evaluate the given model and return the probabilities."""
    model = model[0]
    knowledge = KNOWLEDGE

    handle, tmpfile = tempfile.mkstemp(".pl")
    with open(tmpfile, "w") as f:
        f.write(model)

    handle, outfile = tempfile.mkstemp(".out")

    cmd = ["python", root_path("run_problog.py"), tmpfile, outfile]

    try:
        call_process(cmd, DEFAULT_TIMEOUT, DEFAULT_MEMOUT * (1 << 30))

        with open(outfile) as f:
            result = f.read()
        code, datatype, datavalue = result.split(None, 2)
        return int(code), datatype, datavalue
    except subprocess.CalledProcessError:
        return 500, "text/plain", "ProbLog evaluation exceeded time or memory limit"

    # try :
    #     formula = knowledge.createFrom( PrologString(model) )
    #     result = formula.evaluate()
    #     return 200, 'application/json', json.dumps(result)
    # except Exception as err :
    #     return process_error(err)


@handle_url("/ground")
def run_ground(model):
    """Ground the program given by model and return an SVG of the resulting formula."""
    model = model[0]
    knowledge = LogicFormula

    # from problog.engine import EngineLogger, SimpleEngineLogger
    # EngineLogger.setClass(SimpleEngineLogger)

    try:
        formula = knowledge.createFrom(PrologString(model))

        handle, filename = tempfile.mkstemp(".dot")
        with open(filename, "w") as f:
            f.write(formula.to_dot())
        print(formula)
        result = subprocess.check_output(["dot", "-Tsvg", filename]).decode("utf-8")
        content_type = "application/json"
        # EngineLogger.setClass(None)
        return 200, content_type, json.dumps({"svg": result, "txt": str(formula)})
    except Exception as err:
        # EngineLogger.setClass(None)
        return process_error(err)


def process_error(err):
    """Take the given error raise by ProbLog and produce a meaningful error message."""
    err_type = type(err).__name__
    if err_type == "ParseException":
        return (
            400,
            "text/plain",
            "Parsing error on %s:%s: %s.\n%s"
            % (err.lineno, err.col, err.msg, err.line),
        )
    elif err_type == "UnknownClause":
        return 400, "text/plain", "Predicate undefined: '%s'." % (err)
    elif err_type == "PrologInstantiationError":
        return 400, "text/plain", "Arithmetic operation on uninstantiated variable."
    elif err_type == "UnboundProgramError":
        return 400, "text/plain", "Unbounded program or program too large."
    else:
        traceback.print_exc()
        return 400, "text/plain", "Unknown error: %s" % (err_type)


@handle_url("/models")
def list_models():
    def extract_name(f):
        return os.path.splitext(os.path.basename(f))[0]

    files = map(extract_name, glob.glob(model_path("*.pl")))

    return 200, "application/json", json.dumps(files)


@handle_url("/model")
def load_model(name):
    name = name[0]
    filename = model_path(name + ".pl")
    try:
        with open(filename) as f:
            data = f.read()
        return 200, "text/plain", data
    except Exception:
        return 404, "text/plain", "File not found!"


class ProbLogHTTP(BaseHTTPServer.BaseHTTPRequestHandler):
    def do_POST(self):

        numberOfBytes = int(self.headers["Content-Length"])
        data = self.rfile.read(numberOfBytes)
        query = urlparse.parse_qs(data)
        self.do_GET(query=query)

    def do_GET(self, query=None):
        """Handle a GET request.

        Looks up the URL path in PATHS and executes the corresponding function.
        If the path does not occur in PATHS, treats the path as a filename and serves the file.
        """

        url = urlparse.urlparse(self.path)
        path = url.path
        if query == None:
            query = urlparse.parse_qs(url.query)

        action = PATHS.get(path)
        if action == None:
            self.serveFile(path)
        else:
            code, datatype, data = action(**query)
            self.send_response(code)
            self.send_header("Content-type", datatype)
            self.end_headers()
            if data:
                self.wfile.write(toBytes(data))

    def serveFile(self, filename):
        """Serve a file."""
        if filename == "/":
            filename = "/index.html"

        filetype, encoding = mimetypes.guess_type(filename)
        filename = root_path(filename)
        try:
            with open(filename) as f:
                data = f.read()
            self.send_response(200)
            self.send_header("Content-type", filetype)
            if encoding:
                self.send_header("Content-Encoding", encoding)
            self.end_headers()
            self.wfile.write(toBytes(data))
        except:
            self.send_response(404)
            self.end_headers()
            self.wfile.write(toBytes("File not found!"))


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--port", "-p", type=int, default=DEFAULT_PORT, help="Server listening port"
    )
    parser.add_argument(
        "--timeout",
        "-t",
        type=int,
        default=DEFAULT_TIMEOUT,
        help="Time limit in seconds",
    )
    parser.add_argument(
        "--memout", "-m", type=float, default=DEFAULT_MEMOUT, help="Memory limit in Gb"
    )
    args = parser.parse_args(sys.argv[1:])

    DEFAULT_TIMEOUT = args.timeout
    DEFAULT_MEMOUT = args.memout
    print(
        "Starting server on port %d (timeout=%d, memout=%dGb)"
        % (args.port, DEFAULT_TIMEOUT, DEFAULT_MEMOUT)
    )

    server_address = ("", args.port)
    httpd = BaseHTTPServer.HTTPServer(server_address, ProbLogHTTP)
    httpd.serve_forever()



#! /usr/bin/env python

"""
ProbLog command-line interface.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys
import traceback

from problog import get_evaluatable, register_semiring
from problog.cnf_formula import CNF
from problog.constraint import TrueConstraint
from problog.errors import process_error
from problog.evaluator import Semiring, SemiringProbability
from problog.formula import LogicFormula, LogicDAG
from problog.logic import Term
from problog.maxsat import get_solver, get_available_solvers
from problog.program import PrologFile
from problog.util import init_logger, Timer


def main(argv):
    args = argparser().parse_args(argv)

    if args.use_semiring:
        return main_mpe_semiring(args)
    else:
        return main_mpe_maxsat(args)


def main_mpe_semiring(args):
    inputfile = args.inputfile

    init_logger(args.verbose)

    if args.web:
        result_handler = print_result_json
    else:
        result_handler = print_result

    if args.output is not None:
        outf = open(args.output, "w")
    else:
        outf = sys.stdout

    with Timer("Total"):
        final_result = None
        try:
            pl = PrologFile(inputfile)

            lf = LogicFormula.create_from(pl, label_all=True, avoid_name_clash=True)

            prob, facts = mpe_semiring(lf, args.verbose, minpe=args.minpe)
            final_result = (True, (prob, facts))

        except Exception as err:
            trace = traceback.format_exc()
            err.trace = trace
            final_result = (False, err)
        result_handler(final_result, outf)

        return final_result


def mpe_semiring(lf, verbose=0, solver=None, minpe=False):
    if minpe:
        semiring = SemiringMinPEState()
    else:
        semiring = SemiringMPEState()
    kc_class = get_evaluatable(semiring=semiring)

    if lf.evidence():
        # Query = evidence + constraints
        qn = lf.add_and([y for x, y in lf.evidence()])
        lf.clear_evidence()

        if lf.queries():
            non_atom = []
            atom = []
            qs = []
            for qnm, qi in lf.queries():
                if lf.is_probabilistic(qi):
                    if type(lf.get_node(qi)).__name__ != "atom":
                        non_atom.append(qnm)
                atom.append(qi)
                qs += [qnm, -qnm]
            qs = set(qs)
            if non_atom:
                print(
                    "WARNING: compound queries are not supported in the output: %s"
                    % ", ".join(map(str, non_atom)),
                    file=sys.stderr,
                )

            qn = lf.add_and(
                [lf.add_or((qi, lf.negate(qi)), compact=False) for qi in atom] + [qn]
            )

        else:
            qs = None
        lf.clear_queries()

        query_name = Term("query")
        lf.add_query(query_name, qn, keep_name=True)

        kc = kc_class.create_from(lf)

        results = kc.evaluate(semiring=semiring)
        prob, facts = results[query_name]
        if qs is not None:
            facts &= qs
    else:
        prob, facts = 1.0, []

    return prob, facts


def main_mpe_maxsat(args):
    inputfile = args.inputfile

    if args.web:
        result_handler = print_result_json
    else:
        result_handler = print_result

    if args.output is not None:
        outf = open(args.output, "w")
    else:
        outf = sys.stdout

    with Timer("Total"):
        final_result = None
        try:
            pl = PrologFile(inputfile)

            # filtered_pl = SimpleProgram()
            # has_queries = False
            # for statement in pl:
            #     if 'query/1' in statement.predicates:
            #         has_queries = True
            #     else:
            #         filtered_pl += statement
            # if has_queries:
            #     print('%% WARNING: ignoring queries in file', file=sys.stderr)

            dag = LogicDAG.createFrom(
                pl, avoid_name_clash=True, label_all=True, labels=[("output", 1)]
            )

            prob, output_facts = mpe_maxsat(
                dag, verbose=args.verbose, solver=args.solver, minpe=args.minpe
            )

            final_result = (True, (prob, output_facts))
        except Exception as err:
            trace = traceback.format_exc()
            err.trace = trace
            final_result = (False, err)

        result_handler(final_result, outf)

    if args.output is not None:
        outf.close()

    return final_result


def mpe_maxsat(dag, verbose=0, solver=None, minpe=False):
    logger = init_logger(verbose)
    logger.info("Ground program size: %s" % len(dag))

    cnf = CNF.createFrom(dag, force_atoms=True)
    for qn, qi in cnf.evidence():
        if not cnf.is_true(qi):
            cnf.add_constraint(TrueConstraint(qi))

    queries = list(cnf.labeled())

    logger.info("CNF size: %s" % cnf.clausecount)

    if not cnf.is_trivial():
        solver = get_solver(solver)

        with Timer("Solving"):
            result = frozenset(solver.evaluate(cnf, invert_weights=minpe))
        weights = cnf.extract_weights(SemiringProbability())
        output_facts = None
        prob = 1.0
        if result is not None:
            output_facts = []

            if queries:
                for qn, qi, ql in queries:
                    if qi in result:
                        output_facts.append(qn)
                    elif -qi in result:
                        output_facts.append(-qn)
            for i, n, t in dag:
                if t == "atom":
                    if i in result:
                        if not queries:
                            output_facts.append(n.name)
                        prob *= weights[i][0]
                    elif -i in result:
                        if not queries:
                            output_facts.append(-n.name)
                        prob *= weights[i][1]
    else:
        prob = 1.0
        output_facts = []

    return prob, output_facts


def print_result(result, output=sys.stdout):
    success, result = result
    if success:
        prob, facts = result
        if facts is None:
            print("%% The model is not satisfiable.", file=output)
        else:
            for atom in facts:
                print(atom, file=output)
            if prob is not None:
                print("%% Probability: %.10g" % prob)
        return 0
    else:
        print(process_error(result), file=output)
        return 1


def print_result_json(d, output):
    """Pretty print result.

    :param d: result from run_problog
    :param output: output file
    :param precision:
    :return:
    """
    import json

    result = {}
    success, d = d
    if success:
        prob, facts = d
        result["SUCCESS"] = True
        if facts is not None:
            result["atoms"] = list(
                map(
                    lambda n: (str(-n), False) if n.is_negated() else (str(n), True),
                    facts,
                )
            )
        if prob is not None:
            result["prob"] = round(prob, 10)
    else:
        result["SUCCESS"] = False
        result["err"] = process_error(d)
        result["original"] = str(d)
    print(json.dumps(result), file=output)
    return 0


def reduce_formula(formula, facts):
    # Assume formula is cycle free.

    values = [None] * len(formula)
    for f in facts:
        if f > 0:
            values[f - 1] = True
        else:
            values[-f - 1] = False

    from collections import deque

    nodes = deque(range(1, len(formula) + 1))
    while nodes:
        index = nodes.popleft()
        value = values[index - 1]
        if value is None:
            node = formula.get_node(index)
            nodetype = type(node).__name__
            if nodetype == "atom":
                pass  # Really shouldn't happen
            else:
                children = [
                    values[c] if c > 0 else not values[-c] for c in node.children
                ]
                if nodetype == "disj":
                    if True in children:
                        values[index - 1] = True
                    elif None in children:
                        # not ready yet, push it back on the queue
                        nodes.append(index)
                    else:
                        values[index - 1] = False
                else:
                    if False in children:
                        values[index - 1] = False
                    elif None in children:
                        nodes.append(index)
                    else:
                        values[index - 1] = True
    return values


class SemiringMPEState(Semiring):
    def __init__(self):
        Semiring.__init__(self)

    def zero(self):
        return 0.0, set()

    def one(self):
        return 1.0, set()

    def plus(self, a, b):
        if a[0] > b[0]:
            return a
        elif a[0] < b[0]:
            return b
        else:
            return a[0], a[1]  # | b[1]   # doesn't matter?

    def times(self, a, b):
        return a[0] * b[0], a[1] | b[1]

    def pos_value(self, a, key=None):
        return float(a), {key}

    def neg_value(self, a, key=None):
        return 1.0 - float(a), {-key}

    def is_nsp(self):
        return True

    def result(self, a, formula=None):
        return a

    def ad_complement(self, ws, key=None):
        s = sum([x[0] for x in ws])
        return 1.0 - s, {key}


class SemiringMinPEState(SemiringMPEState):
    def __init__(self):
        SemiringMPEState.__init__(self)

    def plus(self, a, b):
        if a[0] == 0:
            return b
        elif b[0] == 0:
            return a
        elif a[0] > b[0]:
            return b
        elif a[0] < b[0]:
            return a
        else:
            return a[0], a[1]  # | b[1]   # doesn't matter?


def argparser():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("inputfile")
    parser.add_argument(
        "--solver",
        choices=get_available_solvers(),
        default=None,
        help="MaxSAT solver to use",
    )
    parser.add_argument(
        "--full", dest="output_all", action="store_true", help="Also show false atoms."
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=None,
        help="Write output to given file (default: write to stdout)",
    )
    parser.add_argument("--web", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--use-maxsat", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--use-semiring", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--minpe", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("-v", "--verbose", action="count", help="Increase verbosity")
    return parser


if __name__ == "__main__":
    main(sys.argv[1:])



from collections import defaultdict

from problog.extern import problog_export_nondet
from problog.logic import term2list


@problog_export_nondet("+term", "-term", "-list")
def enum_groups(group_values):
    group_values_l = term2list(group_values, False)

    grouped = defaultdict(list)

    for gv in group_values_l:
        grouped[gv.args[0]].append(gv.args[1])
    return list(grouped.items())


@problog_export_nondet("+term", "+term", "-term", "-list")
def enum_groups(groups, values):
    groups_l = term2list(groups, False)
    values_l = term2list(values, False)

    grouped = defaultdict(list)

    for g, v in zip(groups_l, values_l):
        grouped[g].append(v)
    return list(grouped.items())



"""
problog.eval_nodes - Evaluation node classes for the engine_stack
---------------------------------------------------------------------
"""
from problog.engine_builtin import IndirectCallCycleError
from problog.engine_unify import substitute_call_args
from problog.errors import GroundingError
from problog.logic import Term, is_ground, ArithmeticError


class NegativeCycle(GroundingError):
    """The engine does not support negative cycles."""

    def __init__(self, location=None):
        GroundingError.__init__(self, "Negative cycle detected", location)


NODE_TRUE = 0
NODE_FALSE = None


class ResultSet(object):
    def __init__(self):
        self.results = []
        self.index = {}
        self.collapsed = False

    def __setitem__(self, result, node):
        index = self.index.get(result)
        if index is None:
            index = len(self.results)
            self.index[result] = index
            if self.collapsed:
                self.results.append((result, node))
            else:
                self.results.append((result, [node]))
        else:
            assert not self.collapsed
            self.results[index][1].append(node)

    def __getitem__(self, result):
        index = self.index[result]
        result, node = self.results[index]
        return node

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return None

    def keys(self):
        return [result for result, node in self.results]

    def items(self):
        return self.results

    def __len__(self):
        return len(self.results)

    def collapse(self, function):
        if not self.collapsed:
            for i, v in enumerate(self.results):
                result, node = v
                collapsed_node = function(result, node)
                self.results[i] = (result, collapsed_node)
            self.collapsed = True

    def __contains__(self, key):
        return key in self.index

    def __iter__(self):
        return iter(self.results)

    def __str__(self):  # pragma: no cover
        return str(self.results)


def call(obj, args, kwdargs):
    return "e", obj, args, kwdargs


def new_result(obj, result, ground_node, source, is_last):
    return "r", obj, (result, ground_node, source, is_last), {}


def complete(obj, source):
    return "c", obj, (source,), {}


def results_to_actions(
    resultlist,
    engine,
    node,
    context,
    target,
    parent,
    identifier,
    transform,
    is_root,
    database,
    **kwdargs
):
    """Translates a list of results to actions.

    :param results:
    :param node:
    :param context:
    :param target:
    :param parent:
    :param identifier:
    :param transform:
    :param is_root:
    :param database:
    :param kwdargs:
    :return:
    """

    # Output
    actions = []

    n = len(resultlist)
    if n > 0:
        # Transform all the results to result messages.
        for result, target_node in resultlist:
            n -= 1
            if not is_root:
                target_node = engine.propagate_evidence(
                    database, target, node.functor, result, target_node
                )
            if target_node != NODE_FALSE:
                if transform:
                    result = transform(result)
                if result is None:
                    if n == 0:
                        actions += [complete(parent, identifier)]
                else:
                    if (
                        target_node == NODE_TRUE
                        and target.flag("keep_all")
                        and not node.functor.startswith("_problog")
                    ):
                        name = Term(node.functor, *result)
                        target_node = target.add_atom(
                            name, None, None, name=name, source=None
                        )

                    actions += [
                        new_result(parent, result, target_node, identifier, n == 0)
                    ]
            elif n == 0:
                actions += [complete(parent, identifier)]
    else:
        # The goal does not have results: send the completion message.
        actions += [complete(parent, identifier)]
    return actions


class EvalNode(object):
    def __init__(
        self,
        engine,
        database,
        target,
        node_id,
        node,
        context,
        parent,
        pointer,
        identifier=None,
        transform=None,
        call=None,
        current_clause=None,
        include=None,
        exclude=None,
        no_cache=False,
        **extra
    ):
        self.engine = engine
        self.database = database
        self.target = target
        self.node_id = node_id
        self.node = node
        self.context = context
        self.parent = parent
        self.identifier = identifier
        self.pointer = pointer
        self.transform = transform
        self.call = call
        self.on_cycle = False
        self.current_clause = current_clause
        self.include = include
        self.exclude = exclude
        self.no_cache = no_cache

    def notifyResult(self, arguments, node=0, is_last=False, parent=None):
        if parent is None:
            parent = self.parent
        if self.transform:
            arguments = self.transform(arguments)
        if arguments is None:
            if is_last:
                return self.notifyComplete()
            else:
                return []
        else:
            return [new_result(parent, arguments, node, self.identifier, is_last)]

    def notifyComplete(self, parent=None):
        if parent is None:
            parent = self.parent
        return [complete(parent, self.identifier)]

    def createCall(self, node_id, *args, **kwdargs):
        base_args = {}
        base_args["database"] = self.database
        base_args["target"] = self.target
        base_args["context"] = self.context
        base_args["parent"] = self.pointer
        base_args["identifier"] = self.identifier
        base_args["transform"] = None
        base_args["call"] = self.call
        base_args["current_clause"] = self.current_clause
        base_args["no_cache"] = self.no_cache
        base_args["include"] = self.include
        base_args["exclude"] = self.exclude
        base_args.update(kwdargs)
        return call(node_id, args, base_args)

    def createCycle(self):
        self.on_cycle = True
        return []

    def node_str(self):  # pragma: no cover
        return str(self.node)

    def __str__(self):  # pragma: no cover
        if hasattr(self.node, "location"):
            pos = self.database.lineno(self.node.location)
        else:
            pos = None
        if pos is None:
            pos = "??"
        node_type = self.__class__.__name__[4:]
        return "%s %s %s [at %s:%s] | Context: %s" % (
            self.parent,
            node_type,
            self.node_str(),
            pos[0],
            pos[1],
            self.context,
        )


class EvalOr(EvalNode):
    # Has exactly one listener (parent)
    # Has C children.
    # Behaviour:
    # - 'call' creates child nodes and request calls to them
    # - 'call' calls complete if there are no children (C = 0)
    # - 'new_result' is forwarded to parent
    # - 'complete' waits until it is called C times, then sends signal to parent
    # Can be cleanup after 'complete' was sent

    def __init__(self, **parent_args):
        EvalNode.__init__(self, **parent_args)
        self.results = ResultSet()
        if not self.is_buffered():
            self.flushBuffer(True)
        self.to_complete = len(self.node.children)
        self.engine.stats[0] += 1

    def is_buffered(self):
        return not (self.on_cycle or self.engine.unbuffered)

    def isOnCycle(self):
        return self.on_cycle

    def flushBuffer(self, cycle=False):
        func = lambda result, nodes: self.target.add_or(
            nodes, readonly=(not cycle), name=None
        )
        self.results.collapse(func)

    def new_result(self, result, node=NODE_TRUE, source=None, is_last=False):
        if not self.is_buffered():
            res = self.engine._fix_context(result)
            assert self.results.collapsed
            if res in self.results:
                res_node = self.results[res]
                self.target.add_disjunct(res_node, node)
                if is_last:
                    return self.complete(source)
                else:
                    return False, []
            else:
                result_node = self.target.add_or((node,), readonly=False, name=None)
                self.results[res] = result_node
                actions = self.notifyResult(res, result_node)
                if is_last:
                    a, act = self.complete(source)
                    actions += act
                else:
                    a = False
                return a, actions
        else:
            assert not self.results.collapsed
            res = self.engine._fix_context(result)
            self.results[res] = node
            if is_last:
                return self.complete(source)
            else:
                return False, []

    def complete(self, source=None):
        self.to_complete -= 1
        if self.to_complete == 0:
            self.flushBuffer()
            actions = []
            if self.is_buffered():
                for result, node in self.results:
                    actions += self.notifyResult(result, node)
            actions += self.notifyComplete()
            return True, actions
        else:
            return False, []

    def createCycle(self):
        if self.is_buffered():
            self.on_cycle = True
            self.flushBuffer(True)
            actions = []
            for result, node in self.results:
                actions += self.notifyResult(result, node)
            return actions
        else:
            self.on_cycle = True
            return []

    def node_str(self):  # pragma: no cover
        return ""

    def __str__(self):  # pragma: no cover
        return EvalNode.__str__(self) + " tc: " + str(self.to_complete)


class EvalDefine(EvalNode):
    # A buffered Define node.
    def __init__(self, call=None, to_complete=None, is_root=False, **parent_args):
        EvalNode.__init__(self, **parent_args)
        # self.__buffer = defaultdict(list)
        # self.results = None

        self.results = ResultSet()

        self.cycle_children = []
        self.cycle_close = set()
        self.is_cycle_root = False
        self.is_cycle_child = False
        self.is_cycle_parent = False
        self.siblings = []  # These are nodes that have is_cycle_child set

        self.call = (self.node.functor, self.context)
        self.to_complete = to_complete
        self.is_ground = is_ground(*self.context)
        self.is_root = is_root
        self.engine.stats[1] += 1

        if not self.is_buffered():
            self.flushBuffer(True)

    def notifyResult(self, arguments, node=0, is_last=False, parent=None):
        if not self.is_root:
            node = self.engine.propagate_evidence(
                self.database, self.target, self.node.functor, arguments, node
            )
        return super(EvalDefine, self).notifyResult(arguments, node, is_last, parent)

    def notifyResultMe(self, arguments, node=0, is_last=False):
        parent = self.pointer
        return [new_result(parent, arguments, node, self.identifier, is_last)]

    def notifyResultSiblings(self, arguments, node=0, is_last=False):
        parents = self.siblings
        return [
            new_result(parent, arguments, node, self.identifier, is_last)
            for parent in parents
        ]

    def notifyResultChildren(self, arguments, node=0, is_last=False):
        parents = self.cycle_children
        return [
            new_result(parent, arguments, node, self.identifier, is_last)
            for parent in parents
        ]

    def new_result(self, result, node=NODE_TRUE, source=None, is_last=False):
        if self.is_ground and node == NODE_TRUE and not self.target.flag("keep_all"):
            # We have a ground node with a deterministically true proof.
            # We can ignore the remaining proofs.
            # self.engine.ignoring.add(self.pointer)
            pass  # not when there is state
        if self.is_cycle_child:
            assert not self.siblings
            if is_last:
                return True, self.notifyResult(result, node, is_last=is_last)
            else:
                return False, self.notifyResult(result, node, is_last=is_last)
        else:
            if not self.is_buffered() or self.isCycleParent():
                assert self.results.collapsed
                res = self.engine._fix_context(result)
                res_node = self.results.get(res)
                if res_node is not None:
                    self.target.add_disjunct(res_node, node)
                    actions = []
                    if is_last:
                        a, act = self.complete(source)
                        actions += act
                    else:
                        a = False
                    return a, actions
                else:
                    cache_key = self.call  # (self.node.functor, res)
                    if not self.no_cache and cache_key in self.target._cache:
                        # Get direct
                        stored_result = self.target._cache[cache_key]
                        assert len(stored_result) == 1
                        result_node = stored_result[0][1]

                        if not self.is_root:
                            result_node = self.engine.propagate_evidence(
                                self.database,
                                self.target,
                                self.node.functor,
                                res,
                                result_node,
                            )
                    else:
                        if self.engine.label_all:
                            name = Term(self.node.functor, *res)
                        else:
                            name = None
                            if not self.is_root:
                                node = self.engine.propagate_evidence(
                                    self.database,
                                    self.target,
                                    self.node.functor,
                                    res,
                                    node,
                                )
                        result_node = self.target.add_or(
                            (node,), readonly=False, name=name
                        )
                    self.results[res] = result_node
                    if (
                        not self.no_cache
                        and is_ground(*res)
                        and is_ground(*self.call[1])
                    ):
                        self.target._cache[cache_key] = {res: result_node}
                    actions = []
                    # Send results to cycle
                    if not self.is_buffered() and result_node is not NODE_FALSE:
                        actions += self.notifyResult(res, result_node)

                    # TODO what if result_node is NONE?
                    actions += self.notifyResultChildren(
                        res, result_node, is_last=False
                    )
                    actions += self.notifyResultSiblings(
                        res, result_node, is_last=False
                    )
                    # TODO the following optimization doesn't always work, see test/some_cycles.pl
                    # actions += self.notifyResultChildren(res, result_node, is_last=self.is_ground)
                    # if self.is_ground :
                    #     self.engine.cycle_root.cycle_close -= set(self.cycle_children)

                    if is_last:
                        a, act = self.complete(source)
                        actions += act
                    else:
                        a = False
                    return a, actions
            else:
                assert not self.results.collapsed
                res = self.engine._fix_context(result)
                self.results[res] = node
                if is_last:
                    return self.complete(source)
                else:
                    return False, []

    def complete(self, source=None):
        if self.is_cycle_child:
            assert not self.siblings
            return True, self.notifyComplete()
        else:
            self.to_complete -= 1
            if self.to_complete == 0:
                cache_key = self.call
                # cache_key = (self.node.functor, self.context)
                # assert (not cache_key in self.target._cache)
                self.flushBuffer()
                self.target._cache[cache_key] = self.results
                self.target._cache.deactivate(cache_key)
                actions = []
                if self.is_buffered():
                    actions = results_to_actions(self.results, **vars(self))

                    for s in self.siblings:
                        n = len(self.results)
                        if n > 0:
                            for arg, rn in self.results:
                                n -= 1
                                actions.append(
                                    new_result(s, arg, rn, self.identifier, n == 0)
                                )
                        else:
                            actions += [complete(s, self.identifier)]
                else:
                    actions += self.notifyComplete()
                    for s in self.siblings:
                        actions += self.notifyComplete(parent=s)
                return True, actions
            else:
                return False, []

    def flushBuffer(self, cycle=False):
        def func(res, nodes):
            cache_key = self.call
            #            cache_key = (self.node.functor, res)
            if not self.no_cache and cache_key in self.target._cache:
                stored_result = self.target._cache[cache_key]
                if len(stored_result) == 1:
                    node = stored_result[0][1]
                else:
                    if self.engine.label_all:
                        name = Term(self.node.functor, *res)
                    else:
                        name = None
                    node = self.target.add_or([y for x, y in stored_result], name=name)
                # assert (len(stored_result) == 1)
                if not self.is_root:
                    node = self.engine.propagate_evidence(
                        self.database, self.target, self.node.functor, res, node
                    )
            else:
                if self.engine.label_all:
                    name = Term(self.node.functor, *res)
                else:
                    name = None

                if not self.is_root:
                    new_nodes = []
                    for node in nodes:
                        node = self.engine.propagate_evidence(
                            self.database, self.target, self.node.functor, res, node
                        )
                        new_nodes.append(node)
                    nodes = new_nodes
                node = self.target.add_or(nodes, readonly=(not cycle), name=name)
                if not self.no_cache and is_ground(*res) and is_ground(*self.call[1]):
                    self.target._cache[cache_key] = {res: node}

            return node

        self.results.collapse(func)

    def is_buffered(self):
        return not (self.on_cycle or self.engine.unbuffered)

    def isOnCycle(self):
        return self.on_cycle

    def isCycleParent(self):
        return bool(self.cycle_children) or self.is_cycle_parent

    def cycleDetected(self, cycle_parent):
        queue = []
        cycle = self.engine.find_cycle(
            self.pointer, cycle_parent.pointer, cycle_parent.isCycleParent()
        )

        if not cycle:
            cycle_parent.siblings.append(self.pointer)
            self.is_cycle_child = True
            for result, node in cycle_parent.results:
                if type(node) == list:
                    raise IndirectCallCycleError()
                queue += self.notifyResultMe(result, node)
                queue += self.notifyResultMe(result, node)
        else:
            # goal = (self.node.functor, self.context)
            # Get the top node of this cycle.
            cycle_root = self.engine.cycle_root
            # Mark this node as a cycle child
            self.is_cycle_child = True
            # Register this node as a cycle child of cycle_parent
            cycle_parent.cycle_children.append(self.pointer)

            cycle_parent.flushBuffer(True)
            for result, node in cycle_parent.results:
                queue += self.notifyResultMe(result, node)

            if cycle_root is not None and cycle_parent.pointer < cycle_root.pointer:
                # New parent is earlier in call stack as current cycle root
                # Unset current root
                # Unmark current cycle root
                cycle_root.is_cycle_root = False
                # Copy subcycle information from old root to new root
                cycle_parent.cycle_close = cycle_root.cycle_close
                cycle_root.cycle_close = set()
                self.engine.cycle_root = cycle_parent
                queue += cycle_root.createCycle()
                queue += self.engine.notify_cycle(cycle)
                # queue += self.engine.notifyCycle(cycle_root)
                # cycle_root)  # Notify old cycle root up to new cycle root
                cycle_root = None
            if cycle_root is None:  # No active cycle
                # Register cycle root with engine
                self.engine.cycle_root = cycle_parent
                # Mark parent as cycle root
                cycle_parent.is_cycle_root = True
                #
                cycle_parent.cycle_close.add(self.pointer)
                # Queue a create cycle message that will be passed up the call stack.
                queue += self.engine.notify_cycle(cycle)
                # queue += self.engine.notifyCycle(self)
                # Send a close cycle message to the cycle root.
            else:  # A cycle is already active
                # The new one is a subcycle
                # if not self.is_ground :
                cycle_root.cycle_close.add(self.pointer)
                # Queue a create cycle message that will be passed up the call stack.
                queue += self.engine.notify_cycle(cycle)
                if cycle_parent.pointer != self.engine.cycle_root.pointer:
                    to_cycle_root = self.engine.find_cycle(
                        cycle_parent.pointer, self.engine.cycle_root.pointer
                    )
                    if to_cycle_root is None:
                        raise IndirectCallCycleError(
                            self.database.lineno(self.node.location)
                        )
                    queue += cycle_parent.createCycle()
                    queue += self.engine.notify_cycle(to_cycle_root)
        return queue

    def closeCycle(self, toplevel):
        if self.is_cycle_root and toplevel:
            self.engine.cycle_root = None
            actions = []
            for cc in self.cycle_close:
                actions += self.notifyComplete(parent=cc)
            return actions
        else:
            return []

    def createCycle(self):
        if self.on_cycle:  # Already on cycle
            # Pass message to parent
            return []
        elif self.is_cycle_root:
            return []
        elif self.is_buffered():
            # Define node
            self.on_cycle = True
            self.flushBuffer(True)
            actions = []
            for result, node in self.results:
                actions += self.notifyResult(result, node)
                for s in self.siblings:
                    actions += self.notifyResultSiblings(result, node)
            return actions
        else:
            self.on_cycle = True
            return []

    def node_str(self):  # pragma: no cover
        return str(Term(self.node.functor, *self.context))

    def __str__(self):  # pragma: no cover
        extra = ["tc: %s" % self.to_complete]
        if self.is_cycle_child:
            extra.append("CC")
        if self.is_cycle_root:
            extra.append("CR")
        if self.isCycleParent():
            extra.append("CP")
        if self.on_cycle:
            extra.append("*")
        if self.cycle_children:
            extra.append("c_ch: %s" % (self.cycle_children,))
        if self.cycle_close:
            extra.append("c_cl: %s" % (self.cycle_close,))
        if self.siblings:
            extra.append("sbl: %s" % (self.siblings,))
        if not self.is_buffered():
            extra.append("U")
        return EvalNode.__str__(self) + " " + " ".join(extra)


class EvalNot(EvalNode):
    # Has exactly one listener (parent)
    # Has 1 child.
    # Behaviour:
    # - 'new_result' stores results and does not request actions
    # - 'complete: sends out new_results and complete signals
    # Can be cleanup after 'complete' was sent

    def __init__(self, **parent_args):
        EvalNode.__init__(self, **parent_args)
        self.nodes = set()  # Store ground nodes
        self.engine.stats[2] += 1

    def __call__(self):
        return False, [self.createCall(self.node.child)]

    def new_result(self, result, node=NODE_TRUE, source=None, is_last=False):
        if node != NODE_FALSE:
            self.nodes.add(node)
        if is_last:
            return self.complete(source)
        else:
            return False, []

    def complete(self, source=None):
        actions = []
        if self.nodes:
            if self.engine.label_all and self.identifier is not None:
                name = Term('aux_{}'.format(self.identifier))
            else:
                name = None
            or_node = self.target.add_not(self.target.add_or(self.nodes, name=name))
            if or_node != NODE_FALSE:
                actions += self.notifyResult(self.context, or_node)
        else:
            if self.target.flag("keep_all"):
                src_node = self.database.get_node(self.node.child)
                min_var = self.engine.context_min_var(self.context)
                if type(src_node).__name__ == "atom":
                    args, _ = substitute_call_args(
                        src_node.args, self.context, min_var=min_var
                    )
                    name = Term(src_node.functor, *args)
                else:
                    name = None
                node = -self.target.add_atom(
                    name, False, None, name=name, source="negation"
                )
            else:
                node = NODE_TRUE

            actions += self.notifyResult(self.context, node)
        actions += self.notifyComplete()
        return True, actions

    def createCycle(self):
        raise NegativeCycle(location=self.database.lineno(self.node.location))

    def node_str(self):  # pragma: no cover
        return ""


class EvalAnd(EvalNode):
    def __init__(self, **parent_args):
        EvalNode.__init__(self, **parent_args)
        self.to_complete = 1
        self.engine.stats[3] += 1

    def __call__(self):
        return False, [self.createCall(self.node.children[0], identifier=None)]

    def new_result(self, result, node=0, source=None, is_last=False):
        if source is None:  # Result from the first conjunct.
            # We will create a second conjunct, which needs to send a 'complete' signal.
            self.to_complete += 1
            if is_last:
                # Notify self that this conjunct is complete. ('all_complete' will always be False)
                all_complete, complete_actions = self.complete()
                # if False and node == NODE_TRUE :
                #     # TODO THERE IS A BUG HERE
                #     # If there is only one node to complete (the new second conjunct) then
                #     #  we can clean up this node, but then we would lose the ground node of
                #     #  the first conjunct.
                #     # This is ok when it is deterministically true.  TODO make this always ok!
                #     # We can redirect the second conjunct to our parent.
                #     return (self.to_complete==1),
                # [ self.createCall( self.node.children[1], context=result, parent=self.parent ) ]
                # else :
                return (
                    False,
                    [
                        self.createCall(
                            self.node.children[1], context=result, identifier=node
                        )
                    ],
                )
            else:
                # Not the last result: default behaviour
                return (
                    False,
                    [
                        self.createCall(
                            self.node.children[1], context=result, identifier=node
                        )
                    ],
                )
        else:  # Result from the second node
            # Make a ground node
            target_node = self.target.add_and((source, node), name=None)
            if is_last:
                # Notify self of child completion
                all_complete, complete_actions = self.complete()
            else:
                all_complete, complete_actions = False, []
            if all_complete:
                return True, self.notifyResult(result, target_node, is_last=True)
            else:
                return False, self.notifyResult(result, target_node, is_last=False)

    def complete(self, source=None):
        self.to_complete -= 1
        if self.to_complete == 0:
            return True, self.notifyComplete()
        else:
            assert self.to_complete > 0
            return False, []

    def node_str(self):  # pragma: no cover
        return ""

    def __str__(self):  # pragma: no cover
        return EvalNode.__str__(self) + " tc: %s" % self.to_complete


class EvalBuiltIn(EvalNode):
    def __init__(self, call_origin=None, **kwdargs):
        EvalNode.__init__(self, **kwdargs)
        if call_origin is not None:
            self.location = call_origin[1]
        else:
            self.location = None
        self.call_origin = call_origin
        self.engine.stats[4] += 1

    def __call__(self):
        try:
            return self.node(
                *self.context,
                engine=self.engine,
                database=self.database,
                target=self.target,
                location=self.location,
                callback=self,
                transform=self.transform,
                parent=self.parent,
                context=self.context,
                identifier=self.identifier,
                call_origin=self.call_origin,
                current_clause=self.current_clause
            )
        except ArithmeticError as err:
            if self.database and self.location:
                functor = self.call_origin[0].split("/")[0]
                callterm = Term(functor, *self.context)
                base_message = "Error while evaluating %s: %s" % (
                    callterm,
                    err.base_message,
                )
                location = self.database.lineno(self.location)
                raise ArithmeticError(base_message, location)
            else:
                raise err



"""
problog.core - Binary Decision Diagrams
----------------------------------------------

Provides core functionality of ProbLog.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from collections import defaultdict

from .errors import ProbLogError


class ProbLog(object):
    """Static class containing transformation information"""

    def __init__(self):
        raise RuntimeError("This is a static class!")

    transformations = defaultdict(list)
    create_as = defaultdict(list)
    allow_subclass = set()

    @classmethod
    def register_transformation(cls, src, target, action=None):
        """Register a transformation from class src to class target using function action.

        :param src: source function
        :param target: target function
        :param action: transformation function
        """
        cls.transformations[target].append((src, action))

    @classmethod
    def register_create_as(cls, repl, orig):
        """Register that we can create objects of class `repl` in the same way as objects \
        of class `orig`.

        :param repl: object we want to create
        :param orig: object construction we can use instead
        """
        cls.create_as[repl].append(orig)

    @classmethod
    def register_allow_subclass(cls, orig):
        """Register that we can create objects of class `repl` by creating an object of a subclass.

        :param orig:
        """
        cls.allow_subclass.add(orig)

    @classmethod
    def find_paths(cls, src, target, stack=()):
        """Find all possible paths to transform the src object into the target class.

        :param src: object to transform
        :param target: class to tranform the object to
        :param stack: stack of intermediate classes
        :return: list of class, action, class, action, ..., class
        """
        # Create a destination object or any of its subclasses
        if isinstance(src, target):
            yield (target,)
        else:
            targets = [(0, target, target)] + [
                (0, target, d) for d in cls.create_as[target]
            ]
            if target in cls.allow_subclass:
                for d in set(list(cls.transformations) + list(cls.create_as)):
                    if issubclass(d, target) and (
                        not hasattr(d, "is_available") or getattr(d, "is_available")()
                    ):
                        if hasattr(d, "transform_preference"):
                            w = getattr(d, "transform_preference")
                        else:
                            w = 1000
                        targets.append((w, d, d))
            for w, tar, d in sorted(targets, key=lambda v: v[0]):
                for s, action in cls.transformations[d]:
                    if s not in stack:
                        for path in cls.find_paths(src, s, stack + (s,)):
                            yield path + (action, tar)

    @classmethod
    def convert(cls, src, target, **kwdargs):
        """Convert the source object into an object of the target class.

        :param src: source object
        :param target: target class
        :param kwdargs: additional arguments passed to transformation functions
        """

        for d in cls.create_as[target]:
            if type(src) == d:
                return src.clone(target(**kwdargs))

        # Find transformation paths from source to target.
        for path in cls.find_paths(src, target):
            try:
                # A path is a sequence of obj, function, obj/class, ..., obj/class
                current_obj = src
                path = path[1:]
                # Invariant: path[0] is a function, path[1] is an obj/class
                while path:
                    if path[1] is not None:
                        next_obj = path[0](current_obj, path[1](**kwdargs), **kwdargs)
                    else:
                        next_obj = path[1].create_from_default_action(
                            current_obj, **kwdargs
                        )
                    path = path[2:]
                    current_obj = next_obj
                return current_obj
            except TransformationUnavailable:
                # The current transformation strategy didn't work for some reason. Try another one.
                pass
        raise ProbLogError(
            "No conversion strategy found from an object of "
            "class '%s' to an object of class '%s'"
            % (type(src).__name__, getattr(target, "__name__"))
        )


class TransformationUnavailable(Exception):
    """Exception thrown when no valid transformation between two ProbLogObjects can be found."""

    pass


class ProbLogObject(object):
    """Root class for all convertible objects in the ProbLog system."""

    @classmethod
    def create_from(cls, obj, **kwdargs):
        """Transform the given object into an object of the current class using transformations.

        :param obj: obj to transform
        :param kwdargs: additional options
        :return: object of current class
        """
        return ProbLog.convert(obj, cls, **kwdargs)

    # noinspection PyPep8Naming
    @classmethod
    def createFrom(cls, obj, **kwdargs):
        """Transform the given object into an object of the current class using transformations.

        :param obj: obj to transform
        :param kwdargs: additional options
        :return: object of current class
        """
        return cls.create_from(obj, **kwdargs)

    @classmethod
    def create_from_default_action(cls, src):
        """Create object of this class from given source object using default action.

        :param src: source object to transform
        :return: transformed object
        """

        raise ProbLogError("No default conversion strategy defined.")


def transform_create_as(cls1, cls2):
    """Informs the system that cls1 can be used instead of cls2 in any transformations.

    :param cls1:
    :param cls2:
    :return:
    """
    ProbLog.register_create_as(cls1, cls2)


def transform_allow_subclass(cls):
    ProbLog.register_allow_subclass(cls)
    return cls


# noinspection PyPep8Naming
class transform(object):
    """Decorator for registering a transformation between two classes.

    :param cls1: source class
    :param cls2: target class
    :param func: transformation function (for direct use instead of decorator)
    """

    def __init__(self, cls1, cls2, func=None):
        self.cls1 = cls1
        self.cls2 = cls2
        if not issubclass(cls2, ProbLogObject):
            raise TypeError("Conversion only possible for subclasses of ProbLogObject.")
        if func is not None:
            self(func)

    def __call__(self, f):
        ProbLog.register_transformation(self.cls1, self.cls2, f)
        return f


def list_transformations():
    """Print an overview of available transformations."""
    print("Available transformations:")
    for target in ProbLog.transformations:
        print("\tcreate %s.%s" % (target.__module__, target.__name__))
        for src, func in ProbLog.transformations[target]:
            print(
                "\t\tfrom %s.%s by %s.%s"
                % (src.__module__, src.__name__, func.__module__, func.__name__)
            )



"""
Part of the ProbLog distribution.

Copyright 2022 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import re
import subprocess
import unittest

from problog import root_path, system_info


class TestdSharpCompilation(unittest.TestCase):

    def test_dsharp_long_clause(self):
        """
        Bug description: https://github.com/QuMuLab/dsharp/issues/15
        Test dSharp on a CNF with a very long clause (>=66k chars)
        Model count should be exactly 1; and all variables (20058) should be used.
        Previously it erroneously printed only 20056.
        """
        if system_info.get("c2d", False):
            return  # skip if system uses c2d instead of dsharp
        # run dSharp
        filename = root_path("test/specific/", "long_clause.cnf")
        cmd = ["dsharp", filename]
        mc_output = subprocess.run(cmd, stdout=subprocess.PIPE).stdout.decode('utf-8')
        # extract statistics
        used_vars = re.search("used Variables:\t*([0-9]*)\n", mc_output)
        if used_vars is not None:
            expected_used_variables = "20058"
            self.assertEqual(expected_used_variables, used_vars.group(1))



import unittest

from problog.logic import Clause, Term


class TestEquality(unittest.TestCase):
    def test_clause_equality(self):
        c1 = Clause(Term("a"), Term("b"))
        c2 = Clause(Term("c"), Term("b"))
        c3 = Clause(Term("c"), Term("b"))
        self.assertTrue(c2 == c3)
        self.assertFalse(c1 == c2)
        self.assertFalse(c1 == c3)



"""

problog.cnf_formula - CNF
-------------------------

Provides access to CNF and weighted CNF.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from .core import transform
from .evaluator import SemiringLogProbability
from .formula import BaseFormula, LogicDAG
from .util import Timer


class CNF(BaseFormula):
    """A logic formula in Conjunctive Normal Form."""

    # noinspection PyUnusedLocal
    def __init__(self, **kwdargs):
        BaseFormula.__init__(self)
        self._clauses = []  # All clauses in the CNF (incl. comment)
        self._clausecount = 0  # Number of actual clauses (not incl. comment)

    # noinspection PyUnusedLocal
    def add_atom(self, atom, force=False):
        """Add an atom to the CNF.

        :param atom: name of the atom
        :param force: add a clause for each atom to force it's existence in the final CNF
        """
        self._atomcount += 1
        if force:
            self._clauses.append([atom, -atom])
            self._clausecount += 1

    def add_comment(self, comment):
        """Add a comment clause.

        :param comment: text of the comment
        """
        self._clauses.append(["c", comment])

    def add_clause(self, head, body):
        """Add a clause to the CNF.

        :param head: head of the clause (i.e. atom it defines)
        :param body: body of the clause
        """
        self._clauses.append([head] + list(body))
        self._clausecount += 1

    def add_constraint(self, constraint, force=False):
        """Add a constraint.

        :param constraint: constraint to add
        :param force: force constraint to be true even though none of its values are set
        :type constraint: problog.constraint.Constraint
        """
        BaseFormula.add_constraint(self, constraint)
        for c in constraint.as_clauses():
            self.add_clause(force, c)

    def _clause2str(self, clause, weighted=False):
        if weighted:
            raise NotImplementedError()
        else:
            if clause[1] is None:
                return " ".join(map(str, clause[2:])) + " 0"
            else:
                return " ".join(map(str, clause[1:])) + " 0"

    def to_dimacs(
        self,
        partial=False,
        weighted=False,
        semiring=None,
        smart_constraints=False,
        names=False,
        invert_weights=False,
    ):
        """Transform to a string in DIMACS format.

        :param partial: split variables if possibly true / certainly true
        :param weighted: created a weighted (False, :class:`int`, :class:`float`)
        :param semiring: semiring for weight transformation (if weighted)
        :param names: Print names in comments
        :return: string in DIMACS format
        """
        if weighted:
            t = "wcnf"
        else:
            t = "cnf"

        header, content = self._contents(
            partial=partial,
            weighted=weighted,
            semiring=semiring,
            smart_constraints=smart_constraints,
            invert_weights=invert_weights,
        )

        result = "p %s %s\n" % (t, " ".join(map(str, header)))
        if names:
            tpl = "c {{:<{}}} {{}}\n".format(len(str(self._atomcount)) + 1)
            for n, i, l in self.get_names_with_label():
                result += tpl.format(i, n)
        result += "\n".join(map(lambda cl: " ".join(map(str, cl)) + " 0", content))
        return result

    def to_lp(self, partial=False, semiring=None, smart_constraints=False):
        """Transfrom to CPLEX lp format (MIP program).
        This is always weighted.

        :param partial: split variables in possibly true / certainly true
        :param semiring: semiring for weight transformation (if weighted)
        :param smart_constraints: only enforce constraints when variables are set
        :return: string in LP format
        """
        header, content = self._contents(
            partial=partial,
            weighted=False,
            semiring=semiring,
            smart_constraints=smart_constraints,
        )

        if semiring is None:
            semiring = SemiringLogProbability()

        var2str = lambda var: "x%s" % var if var > 0 else "-x%s" % -var

        if partial:
            ct = lambda it: 2 * it
            pt = lambda it: ct(it) - 1

            weights = self.extract_weights(semiring)
            objective = []
            for v in range(0, self.atomcount + 1):
                w_pos, w_neg = weights.get(v, (semiring.one(), semiring.one()))
                if not semiring.is_one(w_pos):
                    w_ct = w_pos
                    objective.append("%s x%s" % (w_ct, ct(v)))
                if not semiring.is_one(w_neg):
                    w_pt = -w_neg
                    objective.append("%s x%s" % (w_pt, pt(v)))
            objective = " + ".join(objective)

        else:
            weights = {}
            for i, w in self.extract_weights(semiring).items():
                w = w[0] - w[1]
                if w != 0:
                    weights[i] = str(w)

            objective = " + ".join(
                [
                    "%s x%s" % (weights[i], i)
                    for i in range(0, self.atomcount + 1)
                    if i in weights
                ]
            )
        result = "maximize\n"
        result += "    obj:" + objective + "\n"
        result += "subject to\n"
        for clause in content:
            n_neg = len([c for c in clause if c < 0])
            result += (
                "    "
                + " + ".join(map(var2str, clause))
                + " >= "
                + str(1 - n_neg)
                + "\n"
            )
        result += "bounds\n"
        for i in range(1, self.atomcount + 1):
            result += "    0 <= x%s <= 1\n" % i
        result += "binary\n"
        result += "    " + " ".join(map(var2str, range(1, self.atomcount + 1))) + "\n"
        result += "end\n"
        return result

    def _contents(
        self,
        partial=False,
        weighted=False,
        semiring=None,
        smart_constraints=False,
        invert_weights=False,
    ):
        # Helper function to determine the certainly true / possibly true names (for partial)

        ct = lambda i: 2 * i
        pt = lambda i: ct(i) - 1
        cpt = lambda i: -pt(-i) if i < 0 else ct(i)

        w_mult = 1
        w_max = []
        weights = None
        if weighted == int:
            w_mult = 10000
            w_min = -10000
            wt1 = lambda w: int(max(w_min, w) * w_mult)
        elif weighted == float:
            w_mult = 1
            w_min = -10000
            wt1 = lambda w: w
        elif weighted:
            w_min = -10000
            w_mult = 10000
            wt1 = lambda w: int(max(w_min, w) * w_mult)
        if weighted:
            if invert_weights:
                wt = lambda w: wt1(-w)
            else:
                wt = wt1
            if semiring is None:
                semiring = SemiringLogProbability()
            weights = self.extract_weights(semiring)

            w_sum = 0.0
            for w_pos, w_neg in weights.values():
                w_sum += max(w_pos, w_min) + max(w_neg, w_min)
            w_max = [int(-w_sum * w_mult) + 1]

        atomcount = self.atomcount
        if partial:
            atomcount *= 2
        clausecount = self.clausecount
        if partial:
            clausecount += atomcount

        clauses = []

        if partial:
            # For each atom: add constraint
            for a in range(1, self.atomcount + 1):
                clauses.append(w_max + [pt(a), -ct(a)])

                if weighted:
                    w_pos, w_neg = weights.get(a, (semiring.one(), semiring.one()))
                    if not semiring.is_one(w_pos):
                        clauses.append([-wt(w_pos), -ct(a)])
                    if not semiring.is_one(w_neg):
                        clauses.append([-wt(w_neg), pt(a)])

            # For each clause:
            for c in self.clauses:
                head, body = c[0], c[1:]
                if type(head) != bool:
                    # Clause does not represent a constraint.
                    head_neg = head < 0
                    head = abs(head)
                    head1, head2 = ct(head), pt(head)
                    if head_neg:
                        head1, head2 = -head1, -head2
                    clauses.append(w_max + [head1, head2] + list(map(cpt, body)))
                elif smart_constraints and not head:
                    # It's a constraint => add an indicator variable.
                    # a \/ -b ===> -pt(a) \/ I  => for all
                    atomcount += 1
                    ind = atomcount
                    v = []
                    for b in body:
                        clauses.append(w_max + [-ct(abs(b)), ind])
                        clauses.append(w_max + [pt(abs(b)), ind])
                        v += [ct(abs(b)), -pt(abs(b))]
                    clauses.append(w_max + v + [-ind])
                    clauses.append(w_max + list(map(cpt, body)) + [-ind])
                else:
                    clauses.append(w_max + list(map(cpt, body)))
        else:
            if weighted:
                for a in range(1, self.atomcount + 1):
                    w_pos, w_neg = weights.get(a, (semiring.one(), semiring.one()))
                    if not semiring.is_one(w_pos):
                        clauses.append([-wt(w_pos), -a])
                    if not semiring.is_one(w_neg):
                        clauses.append([-wt(w_neg), a])
            for c in self.clauses:
                head, body = c[0], c[1:]
                if head is None or type(head) == bool and not head:
                    clauses.append(w_max + list(body))
                else:
                    clauses.append(w_max + [head] + list(body))

        return [atomcount, len(clauses)] + w_max, clauses

    def from_partial(self, atoms):
        """Translates a (complete) conjunction in the partial formula back to the complete formula.

        For example: given an original formula with one atom '1',
         this atom is translated to two atoms '1' (pt) and '2' (ct).

        The possible conjunctions are:

            * [1, 2]    => [1]  certainly true (and possibly true) => true
            * [-1, -2]  => [-1] not possibly true (and certainly true) => false
            * [1, -2]   => []   possibly true but not certainly true => unknown
            * [-1, 2]   => INVALID   certainly true but not possible => invalid (not checked)

        :param atoms: complete list of atoms in partial CNF
        :return: partial list of atoms in full CNF
        """
        result = []
        for s in atoms:
            if s % 2 == 1 and s < 0:
                r = (abs(s) + 1) // 2
                if r in self.get_weights():
                    result.append(-r)
            elif s % 2 == 0 and s > 0:
                r = (abs(s) + 1) // 2
                if r in self.get_weights():
                    result.append(r)
        return result

    def is_trivial(self):
        """Checks whether the CNF is trivial (i.e. contains no clauses)"""
        return self.clausecount == 0

    @property
    def clauses(self):
        """Return the list of clauses"""
        return self._clauses

    @property
    def clausecount(self):
        """Return the number of clauses"""
        return self._clausecount


# noinspection PyUnusedLocal
@transform(LogicDAG, CNF)
def clarks_completion(source, destination, force_atoms=False, **kwdargs):
    """Transform an acyclic propositional program to a CNF using Clark's completion.

    :param source: acyclic program to transform
    :param destination: target CNF
    :param kwdargs: additional options (ignored)
    :return: destination
    """
    with Timer("Clark's completion"):
        # Each rule in the source formula will correspond to an atom.
        num_atoms = len(source)

        # Copy weight information.
        destination.set_weights(source.get_weights())

        # Add atoms.
        for i in range(0, num_atoms):
            destination.add_atom(i + 1, force=force_atoms)

        # Complete other nodes
        # Note: assumes negation is encoded as negative number.
        for index, node, nodetype in source:
            if nodetype == "conj":
                destination.add_clause(index, list(map(lambda x: -x, node.children)))
                for c in node.children:
                    destination.add_clause(-index, [c])
            elif nodetype == "disj":
                destination.add_clause(-index, node.children)
                for c in node.children:
                    destination.add_clause(index, [-c])
            elif nodetype == "atom":
                pass
            else:
                raise ValueError("Unexpected node type: '%s'" % nodetype)

        # Copy constraints.
        for c in source.constraints():
            destination.add_constraint(c)

        # Copy node names.
        for n, i, l in source.get_names_with_label():
            destination.add_name(n, i, l)

        return destination



"""
Module name
"""

import sys


class ProbLogError(Exception):
    """General Problog error. Root of all ProbLog errors that can be blamed on user input.

    :param message: error message
    :param location: location at which the error occurred (referring to an user input file)
    """

    def __init__(self, message, location=None, **extra):
        self.errtype = self.__class__.__name__
        self.base_message = message
        self.location = location
        self.message = self._message()
        for k, v in extra.items():
            setattr(self, k, v)

    def _location_string(self):
        if self.location is None:
            return ""
        if type(self.location) == tuple:
            if len(self.location) != 3:
                return ""
            fn, ln, cn = self.location
            if fn is None:
                return " at %s:%s" % (ln, cn)
            else:
                return " at %s:%s in %s" % (ln, cn, fn)
        else:
            return " at character %s" % self.location

    def _message(self):
        return "%s%s." % (self.base_message, self._location_string())

    def __str__(self):
        return self.message


class ParseError(ProbLogError):
    """Error during parsing."""

    pass


class GroundingError(ProbLogError):
    """Represents an error that occurred during grounding."""

    pass


class CompilationError(ProbLogError):
    """Error during compilation"""

    pass


class InstallError(ProbLogError):
    """Error during installation"""

    pass


class InvalidValue(ProbLogError):
    pass


class UserError(ProbLogError):
    pass


class NonGroundQuery(ProbLogError):
    def __init__(self, query, location):
        super(NonGroundQuery, self).__init__(
            "Query term still contains variables after grounding for query %s" % query,
            location,
        )


class InconsistentEvidenceError(ProbLogError):
    """Error when evidence is inconsistent"""

    def __init__(self, source=None, context=""):
        """

        :param source: evidence term that causes the problem
        :param context: extra message describing the context (e.g. example number in lfi)
        :return:
        """
        self.source = source
        self.context = context
        if source is None:
            ProbLogError.__init__(self, "Inconsistent evidence detected%s" % context)
        else:
            ProbLogError.__init__(
                self, "Inconsistent evidence detected%s: '%s'" % (context, source)
            )


def process_error(err, debug=False):
    """Take the given error raise by ProbLog and produce a meaningful error message.

    :param err: error that was raised
    :param debug: if True, also print original stack trace
    :return: textual representation of the error
    """
    if debug and hasattr(err, "trace"):
        print(err.trace, file=sys.stderr)

    if isinstance(err, ProbLogError):
        return "%s: %s" % (err.__class__.__name__, err)
    elif "Timeout" in str(err) or "timeout" in str(err):
        return "Timeout exceeded"
    elif isinstance(err, KeyboardInterrupt):
        return "Interrupted by user"
    else:
        if not debug and hasattr(err, "trace"):
            print(err.trace, file=sys.stderr)
        return "An unexpected error has occurred."



#! /usr/bin/env python
"""
Query-based sampler for ProbLog 2.1+
===================================

  Concept:
      Each term has a value which is stored in the SampledFormula.
      When an probabilistic atom or choice atom is evaluated:
          - if probability is boolean discrete:
                          determine Yes/No, store in formula and return 0 (True) or None (False)
          - if probability is non-boolean discrete:
                          determine value, store value in formula and return key to value
      Adds builtin sample(X,S) that calls X and returns the sampled value in S.


  How to support evidence?
      => evidence on facts or derived from deterministic data: trivial, just fix value
      => evidence on atoms derived from other probabilistic facts:
          -> query evidence first
          -> if evidence is false, restart immediately
  Currently, evidence is supported through post-processing.


Part of the ProbLog distribution.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import math
import random
import signal
import sys
import time
import traceback

from problog.engine import DefaultEngine, UnknownClause, UnknownClauseInternal
from problog.engine_builtin import check_mode
from problog.engine_unify import UnifyError, unify_value
from problog.errors import process_error, GroundingError
from problog.formula import LogicFormula
from problog.logic import Term, Constant, ArithmeticError, term2list
from problog.program import PrologFile
from problog.util import start_timer, stop_timer, format_dictionary, init_logger

try:
    from tqdm import tqdm
except ImportError:

    class tqdm(object):
        def update(self, x):
            pass


try:

    import numpy.random

    def sample_poisson(l):
        l = [l]
        return numpy.random.poisson(l)[0]


except ImportError:
    numpy = None

    def sample_poisson(l):
        l = list(l)[0]
        # http://www.johndcook.com/blog/2010/06/14/generating-poisson-random-values/
        if l >= 30:
            c = 0.767 - 3.36 / l
            beta = math.pi / math.sqrt(3.0 * l)
            alpha = beta * l
            k = math.log(c) - l - math.log(beta)

            while True:
                u = random.random()
                x = (alpha - math.log((1.0 - u) / u)) / beta
                n = math.floor(x + 0.5)
                if n < 0:
                    continue
                v = random.random()
                y = alpha - beta * x
                lhs = y + math.log(v / (1.0 + math.exp(y)) ** 2)
                rhs = k + n * math.log(l) - math.lgamma(n + 1)
                if lhs <= rhs:
                    return n
        else:
            l2 = math.exp(-l)
            k = 1
            p = 1
            while p > l2:
                k += 1
                p *= random.random()
            return k - 1


class FunctionStore(object):
    def __init__(self, target=None, engine=None, database=None):
        self.target = target
        self.engine = engine
        self.database = database

    # noinspection PyUnusedLocal
    def get(self, key, default=None):
        functor, arity = key

        def _function(*args):
            args = tuple(map(Constant, args))
            term = Term(functor, *args)
            try:
                results = self.engine.call(
                    term, subcall=True, target=self.target, database=self.database
                )
                if len(results) == 1:
                    value_id = results[0][-1]
                    return float(self.target.get_value(value_id))
                elif len(results) == 0:
                    return None
                else:
                    raise ValueError("Unexpected multiple results.")
            except UnknownClause as err:
                sig = err.signature
                raise ArithmeticError("Unknown function %s" % sig)

        return _function


class SampledFormula(LogicFormula):
    def __init__(self, **kwargs):
        LogicFormula.__init__(self, **kwargs)
        self.facts = {}
        self.groups = {}
        self.probability = 1.0  # Try to compute
        self.values = []

        self.distributions = {
            "normal": random.normalvariate,
            "gaussian": random.normalvariate,
            "poisson": sample_poisson,
            "exponential": random.expovariate,
            "beta": random.betavariate,
            "gamma": random.gammavariate,
            "uniform": random.uniform,
            "triangular": random.triangular,
            "vonmises": random.vonmisesvariate,
            "weibull": random.weibullvariate,
            "in_range": random.randint,
        }

    def sample_value(self, term):
        """
        Sample a value from the distribution described by the term.
       :param term: term describing a distribution
       :type term: Term
       :return: value of the term, probability of the choice
       :rtype: Constant
        """
        if term.functor == "fixed":
            return term.args[0], 1.0
        elif term.functor in self.distributions:
            args = map(float, term.args)
            value = self.distributions[term.functor](*args)
            return Constant(value), 0.0
        else:
            raise ValueError("Unknown distribution: '%s'" % term.functor)

    def _is_simple_probability(self, term):
        try:
            float(term)
            return True
        except ArithmeticError:
            return False

    def add_value(self, value):
        self.values.append(value)
        return len(self.values)

    def get_value(self, key):
        if key == 0:
            return None
        return self.values[key - 1]

    def add_atom(
        self,
        identifier,
        probability,
        group=None,
        name=None,
        source=None,
        cr_extra=True,
        is_extra=False,
    ):
        # def add_atom(self, identifier, probability, group=None, name=None, source=None):
        if probability is None:
            return 0

        if group is None:  # Simple fact
            if identifier not in self.facts:
                if self._is_simple_probability(probability):
                    p = random.random()
                    prob = float(probability)
                    value = p < prob
                    if value:
                        result_node = self.TRUE
                        self.probability *= prob
                    else:
                        result_node = self.FALSE
                        self.probability *= 1 - prob
                else:
                    value, prob = self.sample_value(probability)
                    self.probability *= prob
                    result_node = self.add_value(value)
                self.facts[identifier] = result_node
                return result_node
            else:
                return self.facts[identifier]
        else:
            # choice = identifier[-1]
            origin = identifier[:-1]
            if identifier not in self.facts:
                if self._is_simple_probability(probability):
                    p = float(probability)
                    if origin in self.groups:
                        r = self.groups[origin]  # remaining probability in the group
                    else:
                        r = 1.0

                    if r is None or r < 1e-8:
                        # r is too small or another choice was made for this origin
                        value = False
                    else:
                        value = random.random() <= p / r
                    if value:
                        self.probability *= p
                        self.groups[
                            origin
                        ] = None  # Other choices in group are not allowed
                    elif r is not None:
                        self.groups[origin] = r - p  # Adjust remaining probability
                    if value:
                        result_node = self.TRUE
                    else:
                        result_node = self.FALSE
                else:
                    value, prob = self.sample_value(probability)
                    self.probability *= prob
                    result_node = self.add_value(value)
                self.facts[identifier] = result_node
                return result_node
            else:
                return self.facts[identifier]

    def compute_probability(self):
        for k, p in self.groups.items():
            if p is not None:
                self.probability *= p
        self.groups = {}

    def add_and(self, content, **kwdargs):
        i = 0
        for c in content:
            if c is not None and c != 0:
                i += 1
            if i > 1:
                raise ValueError("Can't combine sampled predicates.")
        return LogicFormula.add_and(self, content)

    def add_or(self, content, **kwd):
        i = 0
        for c in content:
            if c is not None and c != 0:
                i += 1
            if i > 1:
                raise ValueError("Bodies for same head should be mutually exclusive.")
        return LogicFormula.add_or(self, content, **kwd)

    def add_not(self, child):
        if child == self.TRUE:
            return self.FALSE
        elif child == self.FALSE:
            return self.TRUE
        else:
            raise ValueError("Can't negate a sampled predicate.")

    def is_probabilistic(self, key):
        """Indicates whether the given node is probabilistic."""
        return False

    # noinspection PyUnusedLocal
    def to_string(
        self,
        db,
        with_facts=False,
        with_probability=False,
        oneline=False,
        as_evidence=False,
        strip_tag=False,
        **extra
    ):
        self.compute_probability()

        if as_evidence:
            base = "evidence(%s)."
        else:
            base = "%s."

        lines = []
        for k, v in self.queries():
            if k.functor.startswith("hidden_"):
                continue
            if strip_tag:
                k = k.args[0]
            if v is not None:
                val = self.get_value(v)
                if val is None:
                    lines.append(base % (str(k)))
                else:
                    if not as_evidence:
                        lines.append("%s = %s." % (str(k), val))
            elif as_evidence:
                lines.append(base % ("\\+" + str(k)))
        if with_facts:
            for k, v in self.facts.items():
                if v == 0:
                    lines.append(base % str(translate(db, k)))
                elif v is None:
                    lines.append(base % ("\\+" + str(translate(db, k))))

        if oneline:
            sep = " "
        else:
            sep = "\n"
        lines = list(set(lines))
        if with_probability:
            lines.append("%% Probability: %.8g" % self.probability)
        return sep.join(lines)

    def to_dict(self):
        result = {}
        for k, v in self.queries():
            if v is not None:
                val = self.get_value(v)
                if val is None:
                    result[k] = True
                else:
                    result[k] = val
            else:
                result[k] = False
        return result


def translate(db, atom_id):
    if type(atom_id) == tuple:
        atom_id, args, choice = atom_id
        return Term("ad_%s_%s" % (atom_id, choice), *args)
    else:
        node = db.get_node(atom_id)
        return Term(node.functor, *node.args)


def builtin_sample(term, result, target=None, engine=None, callback=None, **kwdargs):
    check_mode((term, result), ["cv"], functor="sample")
    # Find the define node for the given query term.
    term_call = term.with_args(*term.args)
    try:
        results = engine.call(term_call, subcall=True, target=target, **kwdargs)
        actions = []
        n = len(term.args)
        for res, node in results:
            res1 = res[:n]
            res_pass = (term.with_args(*res1), target.get_value(node))
            actions += callback.notifyResult(res_pass, 0, False)
        actions += callback.notifyComplete()
        return True, actions
    except UnknownClauseInternal:
        raise UnknownClause(term.signature, term.location)


def builtin_shuffle(lst_in, lst_out, **kwargs):
    check_mode((lst_in, lst_out), ["Lv"], **kwargs)
    # TODO keep correct probability?
    lst_in1 = term2list(lst_in)
    random.shuffle(lst_in1)
    lst_out = term2list(lst_out)
    return [(lst_in, lst_out)]


def builtin_previous(term, default, engine=None, target=None, callback=None, **kwdargs):
    # retrieve term from previous sample, default action if no previous sample
    # See "A. Dries, Declarative data generation with ProbLog, 2015" for more info

    if engine.previous_result is None:
        results = engine.call(default, subcall=True, target=target, **kwdargs)
        actions = []
        for res, node in results:
            try:
                source_values = {}
                result_default = unify_value(default, default(*res), source_values)
                result_term = term.apply(source_values)
                actions += callback.notifyResult(
                    (result_term, result_default), node, False
                )
            except UnifyError:
                pass
        actions += callback.notifyComplete()

    else:
        actions = []
        for n, i in engine.previous_result.queries():
            try:
                source_values = {}
                result_term = unify_value(term, n, source_values)
                result_default = default.apply(source_values)
                if i == 0 or i is None:
                    v = i
                else:
                    v = target.add_value(engine.previous_result.get_value(i))
                actions += callback.notifyResult(
                    (result_term, result_default), v, False
                )
            except UnifyError:
                pass
        actions += callback.notifyComplete()
    return True, actions


def ground(engine, db, target):
    db = engine.prepare(db)
    # Load queries: use argument if available, otherwise load from database.
    queries = [q[0] for q in engine.query(db, Term("query", None))]
    evidence = engine.query(db, Term("evidence", None, None))
    evidence += engine.query(db, Term("evidence", None))

    for query in queries:
        if not isinstance(query, Term):
            raise GroundingError("Invalid query")  # TODO can we add a location?

    for ev in evidence:
        if not isinstance(ev[0], Term):
            raise GroundingError("Invalid evidence")  # TODO can we add a location?

    # Ground queries
    queries = [(target.LABEL_QUERY, q) for q in queries]
    engine.ground_queries(db, target, queries)
    return target


def init_engine(**kwdargs):
    engine = DefaultEngine(**kwdargs)
    engine.add_builtin("sample", 2, builtin_sample)
    engine.add_builtin("value", 2, builtin_sample)
    engine.add_builtin("previous", 2, builtin_previous)
    engine.previous_result = None
    return engine


def init_db(engine, model, propagate_evidence=False):
    db = engine.prepare(model)

    if propagate_evidence:
        evidence = engine.query(db, Term("evidence", None, None))
        evidence += engine.query(db, Term("evidence", None))

        ev_target = LogicFormula()
        engine.ground_evidence(db, ev_target, evidence)
        ev_target.lookup_evidence = {}
        ev_nodes = [
            node
            for name, node in ev_target.evidence()
            if node != 0 and node is not None
        ]
        ev_target.propagate(ev_nodes, ev_target.lookup_evidence)

        evidence_facts = []
        for index, value in ev_target.lookup_evidence.items():
            node = ev_target.get_node(index)
            if ev_target.is_true(value):
                evidence_facts.append((node[0], 1.0) + node[2:])
            elif ev_target.is_false(value):
                evidence_facts.append((node[0], 0.0) + node[2:])
    else:
        evidence_facts = []
        ev_target = None

    return db, evidence_facts, ev_target


class RateCounter(object):
    def __init__(self):
        self.last_tick = None
        self.last_count = 0
        self.counts = []
        self.rate = tqdm()

    def update(self):
        t = time.time()
        if t != self.last_tick:
            if self.last_tick is not None:
                self.counts.append((self.last_tick, self.last_count))
                self.last_count = 0
                self.last_tick = t
        self.last_count += 1
        self.rate.update(1)


def sample(
    model,
    n=1,
    format="str",
    propagate_evidence=False,
    distributions=None,
    progress=False,
    **kwdargs
):
    engine = init_engine(**kwdargs)
    db, evidence, ev_target = init_db(engine, model, propagate_evidence)
    i = 0
    r = 0

    if progress:
        rate = RateCounter()

    try:
        while i < n or n == 0:
            target = SampledFormula()
            if distributions is not None:
                target.distributions.update(distributions)

            for ev_fact in evidence:
                target.add_atom(*ev_fact)

            engine.functions = FunctionStore(target=target, database=db, engine=engine)
            result = ground(engine, db, target=target)
            if verify_evidence(engine, db, ev_target, target):
                if format == "str":
                    yield result.to_string(db, **kwdargs)
                else:
                    yield result.to_dict()
                i += 1
            else:
                r += 1
            engine.previous_result = result
            if progress:
                rate.update()
    except KeyboardInterrupt:
        pass
    if r:
        logging.getLogger("problog_sample").info("Rejected samples: %s" % r)


def verify_evidence(engine, db, ev_target, q_target):
    if ev_target is None:
        evidence = engine.query(db, Term("evidence", None, None))
        evidence += engine.query(db, Term("evidence", None))

        engine.ground_evidence(db, q_target, evidence)
        for name, node in q_target.evidence():
            if q_target.is_false(node):
                return False
        return True
    else:
        weights = {}
        for k, v in q_target.facts.items():
            p = 1.0 if q_target.is_true(v) else 0.0
            i = ev_target.add_atom(k, p)
            weights[i] = p
            weights[-i] = 1.0 - p

        for k, n, t in ev_target:
            if t == "atom" and n.group is not None and not k in weights:
                if q_target.groups.get(n.group, 1.0) is None:
                    p = 0.0
                else:
                    p = 1.0
                weights[k] = p
                weights[-k] = 1.0 - p

        relevant_nodes = set(
            [x for x, y in enumerate(ev_target.extract_relevant()) if y]
        )

        update = True
        while relevant_nodes and update:
            update = False
            next_iteration = set()
            while relevant_nodes:
                i = relevant_nodes.pop()
                if i in weights:
                    w = weights[i]
                else:
                    n = ev_target.get_node(i)
                    t = type(n).__name__
                    if t == "atom":
                        w = float(n.probability)
                        weights[i] = w
                        weights[-i] = 1.0 - w
                        update = True
                    elif t == "disj":
                        cn = [weights[c] for c in n.children if c in weights]
                        if cn:
                            w = max(cn)
                        else:
                            w = 0.0
                        if w > 0.0:
                            weights[i] = w
                            weights[-i] = 1.0 - w
                            update = True
                        elif len(cn) != len(n.children):
                            next_iteration.add(i)
                        else:
                            weights[i] = w
                            weights[-i] = 1.0 - w
                            update = True

                    elif t == "conj":
                        cn = [weights[c] for c in n.children if c in weights]
                        if cn:
                            w = min(cn)
                        else:
                            w = 1.0

                        if w == 0.0:
                            weights[i] = w
                            update = True
                        elif len(cn) != len(n.children):
                            next_iteration.add(i)
                        else:
                            weights[i] = w
                            update = True
            relevant_nodes = next_iteration

        for e, v in ev_target.evidence():
            if weights.get(v, 0.0) == 0.0:
                return False
        return True


# noinspection PyUnusedLocal
def estimate(model, n=0, propagate_evidence=False, **kwdargs):
    from collections import defaultdict

    engine = init_engine(**kwdargs)
    db, evidence, ev_target = init_db(engine, model, propagate_evidence)

    start_time = time.time()
    estimates = defaultdict(float)
    counts = 0.0
    r = 0
    try:
        while n == 0 or counts < n:
            target = SampledFormula()
            for ev_fact in evidence:
                target.add_atom(*ev_fact)

            result = ground(engine, db, target=target)
            if verify_evidence(engine, db, ev_target, target):
                for k, v in result.queries():
                    if v == 0:
                        estimates[k] += 1.0
                counts += 1.0
            else:
                r += 1
            engine.previous_result = result
    except KeyboardInterrupt:
        pass
    except SystemExit:
        pass

    total_time = time.time() - start_time
    rate = counts / total_time
    print(
        "%% Probability estimate after %d samples (%.4f samples/second):"
        % (counts, rate)
    )

    if r:
        logging.getLogger("problog_sample").info("Rejected samples: %s" % r)

    for k in estimates:
        estimates[k] = estimates[k] / counts
    return estimates


def print_result(result, output=sys.stdout, oneline=False):
    success, result = result
    if success:
        # result is a sequence of samples
        first = True
        for s in result:
            if not oneline and not first:
                print("----------------", file=output)
            first = False
            print(s, file=output)
    else:
        print(process_error(result), file=output)


def print_result_json(d, output, **kwdargs):
    """Pretty print result.

    :param d: result from run_problog
    :param output: output file
    :param precision:
    :return:
    """
    import json

    result = {}
    success, d = d
    if success:
        result["SUCCESS"] = True
        result["results"] = [[(str(k), str(v)) for k, v in dc.items()] for dc in d]
    else:
        result["SUCCESS"] = False
        result["test"] = str(type(d))
        result["err"] = process_error(d)
        result["original"] = str(d)
    print(json.dumps(result), file=output)
    return 0


def get_argument_parser():
    import argparse

    parser = argparse.ArgumentParser()

    parser.add_argument("filename")
    parser.add_argument(
        "-N",
        "-n",
        type=int,
        dest="n",
        default=argparse.SUPPRESS,
        help="Number of samples.",
    )
    parser.add_argument(
        "--with-facts",
        action="store_true",
        help="Also output choice facts (default: just queries).",
    )
    parser.add_argument(
        "--with-probability", action="store_true", help="Show probability."
    )
    parser.add_argument(
        "--as-evidence", action="store_true", help="Output as evidence."
    )
    parser.add_argument(
        "--propagate-evidence",
        dest="propagate_evidence",
        default=False,
        action="store_true",
        help="Enable evidence propagation",
    )
    parser.add_argument(
        "--dont-propagate-evidence",
        action="store_false",
        dest="propagate_evidence",
        default=False,
        help="Disable evidence propagation",
    )
    parser.add_argument(
        "--oneline", action="store_true", help="Format samples on one line."
    )
    parser.add_argument(
        "--estimate",
        action="store_true",
        help="Estimate probability of queries from samples.",
    )
    parser.add_argument(
        "--timeout",
        "-t",
        type=int,
        default=0,
        help="Set timeout (in seconds, default=off).",
    )
    parser.add_argument(
        "--output", "-o", type=str, default=None, help="Filename of output file."
    )
    parser.add_argument("--web", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--verbose", "-v", action="count", help="Verbose output")
    parser.add_argument("--seed", "-s", type=float, help="Random seed", default=None)
    parser.add_argument("--full-trace", action="store_true")
    parser.add_argument(
        "--strip-tag", action="store_true", help="Strip outermost tag from output."
    )
    parser.add_argument(
        "-a",
        "--arg",
        dest="args",
        action="append",
        help="Pass additional arguments to the cmd_args builtin.",
    )
    parser.add_argument("--progress", help="show progress", action="store_true")
    return parser


def main(args, result_handler=None):
    parser = get_argument_parser()
    args = parser.parse_args(args)

    init_logger(args.verbose, "problog_sample")

    if args.seed is not None:
        random.seed(args.seed)
    else:
        seed = random.random()
        logging.getLogger("problog_sample").debug("Seed: %s", seed)
        random.seed(seed)

    pl = PrologFile(args.filename)

    outf = sys.stdout
    if args.output is not None:
        outf = open(args.output, "w")

    if args.timeout:
        start_timer(args.timeout)

    # noinspection PyUnusedLocal
    def signal_term_handler(*sigargs):
        sys.exit(143)

    signal.signal(signal.SIGTERM, signal_term_handler)

    if result_handler is not None or args.web:
        outformat = "dict"
        if result_handler is None:
            result_handler = print_result_json
    else:
        outformat = "str"
        result_handler = print_result

    final_result = None
    try:
        if args.estimate:
            results = estimate(pl, **vars(args))
            print(format_dictionary(results))
        else:
            final_result = (True, sample(pl, format=outformat, **vars(args)))
            result_handler(final_result, output=outf, oneline=args.oneline)
    except Exception as err:
        trace = traceback.format_exc()
        err.trace = trace
        final_result = (False, err)
        result_handler(final_result, output=outf)

    if args.timeout:
        stop_timer()

    if args.output is not None:
        outf.close()

    return final_result


if __name__ == "__main__":
    main(sys.argv[1:])



import os
import random
import unittest
from pathlib import Path

from problog.logic import Term

from problog.program import PrologString
from problog.tasks import sample

dirname = os.path.dirname(__file__)
test_folder = Path(dirname, "./../../test/sample")


class TestSampleTask(unittest.TestCase):
    def setUp(self) -> None:
        random.seed(12345)

    def get_samples(self, file_name, num_samples=1, propagate_evidence=False):
        file_name = test_folder / file_name
        model = PrologString(file_name.read_text())
        sample_generator = sample.sample(
            model, n=num_samples, format="dict", propagate_evidence=propagate_evidence
        )
        result_list = list(sample_generator)
        self.assertEqual(
            num_samples,
            len(result_list),
            "Was not able to sample the right number of samples from " + str(file_name),
        )
        return result_list

    @staticmethod
    def estimate(file_name, num_samples=1, propagate_evidence=False):
        file_name = test_folder / file_name
        model = PrologString(file_name.read_text())
        return sample.estimate(
            model, n=num_samples, format="dict", propagate_evidence=propagate_evidence
        )

    @staticmethod
    def count_number_of_atoms(sample_list, term):
        return len(
            [
                sampled_dict
                for sampled_dict in sample_list
                if term in sampled_dict and sampled_dict[term] is True
            ]
        )

    def test_some_heads_distribution(self):
        samples = self.get_samples("some_heads.pl", num_samples=1000)
        self.assertAlmostEqual(
            800, self.count_number_of_atoms(samples, Term("someHeads")), delta=50
        )

    def test_some_heads_evidence(self):
        samples = self.get_samples("some_heads_evidence.pl", num_samples=100)
        number_of_some_heads = self.count_number_of_atoms(samples, Term("someHeads"))
        self.assertAlmostEqual(60, number_of_some_heads, delta=10)

    def test_some_heads_propagate_evidence(self):
        samples = self.get_samples(
            "some_heads_evidence.pl", num_samples=1000, propagate_evidence=True
        )
        number_of_some_heads = self.count_number_of_atoms(samples, Term("someHeads"))
        self.assertAlmostEqual(600, number_of_some_heads, delta=50)

    def test_some_heads_estimate(self):
        with_propagation = self.estimate(
            "some_heads.pl", num_samples=1000, propagate_evidence=False
        )
        self.assertAlmostEqual(0.8, with_propagation[Term("someHeads")], delta=0.05)
        without_propagation = self.estimate(
            "some_heads.pl", num_samples=1000, propagate_evidence=True
        )
        self.assertAlmostEqual(0.8, without_propagation[Term("someHeads")], delta=0.05)

    def test_some_heads_previous(self):
        samples = self.get_samples(
            "some_heads_previous.pl", num_samples=100, propagate_evidence=True
        )
        previous = False
        for s in samples:
            self.assertEqual(
                s[Term("previous", Term("someHeads"), Term("false"))], previous
            )
            previous = s[Term("someHeads")]

    def test_maze_generator(self):
        samples = self.get_samples("maze.pl", num_samples=1)
        self.assertEqual(1, len(samples))
        connected_predicates = [
            p for p in samples[0].keys() if "connected" in p.functor
        ]
        self.assertTrue(25 < len(connected_predicates))

    def test_heights(self):
        samples = self.get_samples("heights.pl", num_samples=10)
        self.assertEqual(10, len(samples))
        for s in samples:
            self.assertAlmostEqual(
                360, float(s[Term("twice_height", Term("harry"))]), delta=60
            )
        print(samples)

    # Tasks

    def test_some_heads_task(self):
        file_name = test_folder / "some_heads.pl"
        result = sample.main([str(file_name), "-n", str(1)])
        success, data = result
        if not success:
            self.fail("Could not successfully sample" + str(file_name))

    def test_some_heads_task_estimate(self):
        file_name = test_folder / "some_heads.pl"
        sample.main([str(file_name), "-n", str(100), "--estimate"])

    def test_some_heads_task_propagate_evidence(self):
        file_name = test_folder / "some_heads_evidence.pl"
        result = sample.main([str(file_name), "-n", str(10), "--propagate-evidence"])
        success, data = result
        if not success:
            self.fail("Could not successfully sample" + str(file_name))



"""
Part of the ProbLog distribution.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import glob
import os
import sys
import unittest

from problog.formula import LogicDAG
from problog.program import PrologFile, DefaultPrologParser, ExtendedPrologFactory
from problog.tasks.bayesnet import formula_to_bn

if __name__ == "__main__":
    sys.path.insert(
        0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../../"))
    )

from problog import root_path


class TestBNGeneric(unittest.TestCase):
    pass


def read_expected_result(filename):
    """
    Get the expected result from the given file. The expected outcome is the % message following after
    %Expected outcome:
    :param filename: The name of the file
    :return: The expected result.strip().
    :rtype: str
    """
    result = ""
    with open(filename) as f:
        reading = False
        for l in f:
            l = l.strip()
            if l.startswith("%Expected outcome:"):
                reading = True
            elif reading:
                if l.lower().startswith("% error"):
                    return l[len("% error") :].strip()
                elif l.startswith("%"):
                    result = result + "\n" + l[1:]
                else:
                    reading = False
    return result.strip()


def createBNTestGeneric(filename, logspace=False):
    correct = read_expected_result(filename)

    def test(self):
        try:
            target = LogicDAG  # Break cycles
            gp = target.createFrom(
                PrologFile(
                    filename, parser=DefaultPrologParser(ExtendedPrologFactory())
                ),
                label_all=True,
                avoid_name_clash=False,
                keep_order=True,  # Necessary for to prolog
                keep_duplicates=False,
            )
            bn = formula_to_bn(gp)
            computed = str(bn).strip()
        except Exception as err:
            e = err
            computed = None

        if computed is None:
            self.assertEqual(correct, type(e).__name__)
        else:
            self.assertIsInstance(computed, str)
            self.assertEqual(correct, computed)

    return test


if __name__ == "__main__":
    filenames = sys.argv[1:]
else:
    filenames = glob.glob(root_path("test/bn", "*.pl"))

for testfile in filenames:
    testname = "test_bn_" + os.path.splitext(os.path.basename(testfile))[0]
    setattr(TestBNGeneric, testname, createBNTestGeneric(testfile, True))

if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestBNGeneric)
    unittest.TextTestRunner(verbosity=2).run(suite)



"""
Part of the ProbLog distribution.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import sys
import time
from collections import defaultdict

from .logic import Term


def printtrace(func):
    def _wrapped_function(*args, **kwdargs):
        print("CALL:", func.__name__, args, kwdargs)
        result = func(*args, **kwdargs)
        print("RETURN:", func.__name__, args, kwdargs, result)
        return result

    return _wrapped_function


def printtrace_self(func):
    def _wrapped_function(self, *args, **kwdargs):
        print("CALL:", func.__name__, args, kwdargs)
        result = func(self, *args, **kwdargs)
        print("RETURN:", func.__name__, args, kwdargs, result)
        return result

    return _wrapped_function


class EngineTracer(object):
    def __init__(self, keep_trace=True, interactive=False):
        self.call_redirect = {}
        self.call_results = defaultdict(int)
        self.level = 0  # level of calls (depth of stack)
        self.stack = []  # stack of calls (approximate?)
        self.interactive = interactive
        if keep_trace:
            self.trace = []
        else:
            self.trace = None

        self.time_start = {}
        self.timestats = defaultdict(float)
        self.resultstats = defaultdict(int)
        self.callstats = defaultdict(int)
        self.time_start_global = None

    def interact(self, record):
        if self.interactive and record:
            print(self.show_record(record), end="\t?")
            a = sys.stdin.readline()
            if a.strip() == "gp":
                pass
            elif a.strip() == "l":
                self.interactive = False

    def process_message(self, msgtype, msgtarget, msgargs, context):
        if msgtarget is None and msgtype == "r" and msgtarget in self.call_redirect:
            self.call_result(*(self.call_redirect[msgtarget] + (msgargs[0],)))

        if msgtype == "r" and msgargs[3] and msgtarget in self.call_redirect:
            self.call_return(*self.call_redirect[msgtarget])
            del self.call_redirect[msgtarget]
        elif msgtype == "c" and msgtarget in self.call_redirect:
            self.call_return(*self.call_redirect[msgtarget])
            del self.call_redirect[msgtarget]

    def call_create(self, node_id, functor, context, parent, location=None):
        if self.time_start_global is None:
            self.time_start_global = time.time()

        term = Term(functor, *context, location=location)
        record = (self.level, "call", term, time.time() - self.time_start_global)
        if self.trace is not None:
            self.trace.append(record)
        self.time_start[term] = time.time(), location
        self.stack.append(term)
        self.level += 1
        self.call_redirect[parent] = (node_id, functor, context)
        self.interact(record)

    def call_result(self, node_id, functor, context, result=None, location=None):
        term = Term(functor, *context, location=location)
        record = (
            self.level,
            "result",
            term,
            time.time() - self.time_start_global,
            result,
        )
        if self.trace is not None:
            self.trace.append(record)
        self.call_results[(node_id, term)] += 1
        self.interact(record)

    def call_return(self, node_id, functor, context, location=None):
        term = Term(functor, *context, location=location)
        self.level -= 1
        ts, loc = self.time_start[term]

        now = time.time()
        if self.stack:
            self.stack.pop(-1)
        if self.call_results[(node_id, term)] > 0:
            record = (
                self.level,
                "complete",
                term,
                now - self.time_start_global,
                now - ts,
            )
        else:
            record = (self.level, "fail", term, now - self.time_start_global, now - ts)
        if self.trace is not None:
            self.trace.append(record)

        self.timestats[(term, loc)] += now - ts
        self.resultstats[(term, loc)] = self.call_results[node_id, term]
        self.callstats[(term, loc)] += 1
        # self.call_results[(node_id, term)] = 0
        self.interact(record)

    def show_profile(self, aggregate=0):
        """Creates a table with profile information.

        :param aggregate: aggregation level (0: no aggregation, 1: same call, 2: predicate)
        :return: string
        """
        s = ""
        if aggregate == 0:
            s += "%50s\t %7s \t %4s \t %4s \t %s \n" % (
                "call",
                "time",
                "#sol",
                "#call",
                "location",
            )
            s += "-" * 100 + "\n"
            for tm, key in sorted(((t, k) for k, t in self.timestats.items()), key=lambda t: t[0]):
                term, location = key
                nb = self.resultstats[key]
                cl = self.callstats[key]
                location = location_string(location)
                s += "%50s\t %.5f \t %d \t %d \t [%s]\n" % (term, tm, nb, cl, location)
        else:
            timestats_agg = defaultdict(float)
            resultstats_agg = defaultdict(int)
            callstats_agg = defaultdict(int)

            for k, t in self.timestats.items():
                if aggregate == 1:
                    key = k[0]
                else:
                    key = k[0].signature
                timestats_agg[key] += t
                resultstats_agg[key] += self.resultstats[k]
                callstats_agg[key] += self.callstats[k]

            s += "%50s\t %7s \t %4s \t %4s\n" % ("call", "time", "#sol", "#call")
            s += "-" * 100 + "\n"
            for tm, key in sorted((t, k) for k, t in timestats_agg.items()):
                nb = resultstats_agg[key]
                cl = callstats_agg[key]
                s += "%50s\t %.5f \t %d \t %d\n" % (key, tm, nb, cl)

        return s

    def show_trace(self):
        s = ""
        if self.trace is not None:
            for record in self.trace:
                s += self.show_record(record) + "\n"
        return s

    def show_record(self, record):
        lvl = record[0]
        msg = record[1]
        term = record[2]
        tm_cumul = record[3]
        if msg == "call":
            s = "%s %s %s {%.5f} [%s]" % (
                " " * lvl,
                msg,
                term,
                tm_cumul,
                location_string(term.location),
            )
        elif msg == "result":
            args = record[4]
            s = "%s %s %s %s {%.5f} [%s]" % (
                " " * lvl,
                msg,
                term,
                args,
                tm_cumul,
                location_string(term.location),
            )
        else:
            tm_local = record[4]
            s = "%s %s %s {%.5f} {%.5f} [%s]" % (
                " " * lvl,
                msg,
                term,
                tm_cumul,
                tm_local,
                location_string(term.location),
            )
        return s


def location_string(location):
    if location is None:
        return ""
    if type(location) == tuple:
        fn, ln, cn = location
        if fn is None:
            return "at %s:%s" % (ln, cn)
        else:
            return "at %s:%s in %s" % (ln, cn, fn)
    else:
        return "at character %s" % location


# Assume the program
#   a(1).
#   a(2).
#   p(X,Y) :- a(X), a(Y), X \= Y.
#   p(1,1).

# Call stack:
#   -> p(X,Y) == [1] a(X), [2] a(Y), [3] X \= Y
#       [1] a(X), a(Y), X \= Y
#           -> a(X)
#           <- a(1) [FACT]
#           <- a(2) [FACT]
#       [2] a(Y), 1\=X
#           -> a(Y) [CACHE]
#           <- a(1)
#           <- a(2)
#       [3] 1 \= 1
#           -> 1 \= 1
#           <- FAIL
#       [3] 1 \= 2
#           -> 1 \= 2
#           <- TRUE
#       [2] a(Y), 2 \= Y
#           -> a(Y) [CACHE]
#           <- a(1)
#           <- a(2)
#       [3] 2 \= 1
#           -> 2 \= 1
#           <- TRUE
#       [3] 2 \= 2
#           -> 2 \= 2
#           <- FAIL
#   <- p(1,2)
#   <- p(2,1)
#   <- p(1,1) [FACT]



"""
test_lfi.py - Test class for LFI problems
"""
import random

from problog import root_path
import unittest
import os
import sys
import glob
from problog.learning.lfi import lfi_wrapper, LFIProblem

try:
    from pysdd import sdd

    has_sdd = True
except Exception as err:
    print("SDD library not available due to error: ", err, file=sys.stderr)
    has_sdd = False


class TestLFI(unittest.TestCase):
    def setUp(self):
        try:
            self.assertSequenceEqual = self.assertItemsEqual
        except AttributeError:
            self.assertSequenceEqual = self.assertCountEqual


def read_result(filename):
    results = []
    with open(filename) as f:
        reading = False
        for l in f:
            l = l.strip()
            if l.startswith("%Expected outcome:"):
                reading = True
            elif reading:
                if l.lower().startswith("% error: "):
                    return l[len("% error: "):].strip()
                elif l.startswith("% "):
                    res = l[2:]
                    results.append(res)
                else:
                    reading = False
    return results


def createTestLFI(filename, evaluatables):
    def test(self):
        for eval_name in evaluatables:
            for logspace in [False, True]:
                with self.subTest(evaluatable=eval_name, logspace=logspace):
                    test_func(self, evaluatable=eval_name, logspace=logspace)

    def test_func(self, evaluatable, logspace):
        model = filename
        examples = filename.replace(".pl", ".ev")
        expectedlines = read_result(model)

        if not os.path.exists(examples):
            raise Exception("Evidence file is missing: {}".format(examples))

        try:
            d = {
                "max_iter": 10000,
                "min_improv": 1e-10,
                "leakprob": None,
                "propagate_evidence": False,
                "logspace": logspace,
                "normalize": True,
                "web": False,
                "args": None,
                "infer_AD_values": True
            }
            random.seed(a=1000)
            score, weights, names, iterations, lfi = lfi_wrapper(
                model, [examples], evaluatable, d
            )
            outlines = lfi.get_model()
        except Exception as _:
            assert expectedlines == "NonGroundProbabilisticClause"
            return

        outlines = outlines.split("\n")[:-1]
        assert len(expectedlines) == len(outlines)

        # We need to sort outlines and expectedlines, because apparently the order of the rules can
        # depend on the target language (-k flag).
        outlines.sort(key=lambda x: x if "::" not in x else x.split("::")[1])
        expectedlines.sort(key=lambda x: x if "::" not in x else x.split("::")[1])

        # Compare expected program and learned program line by line
        for expectedline, outline in zip(expectedlines, outlines):
            # When there are probabilities
            if "::" in outline:
                # Break the lines into components where each component has exactly one probability
                expectedline_comps = expectedline.split(";")
                outline_comps = outline.split(";")
                new_expectedline_comps = []
                new_outline_comps = []
                assert len(outline_comps) == len(expectedline_comps)
                # Compare one expected probability and one learned probability at a time
                for expectedline_comp, outline_comp in zip(
                        expectedline_comps, outline_comps
                ):
                    outline_comp = outline_comp.strip()
                    expectedline_comp = expectedline_comp.strip()
                    # When the learned prob in outline_component does not matter,
                    # discard the learned probability
                    if "<RAND>" in expectedline_comp:
                        outline_comp = "<RAND>::" + outline_comp.split("::")[1]
                    else:
                        # Round the expected and learned probabilities
                        rounded_outline_comp_prob = "{:.6f}".format(
                            float(outline_comp.split("::")[0])
                        )
                        rounded_expectedline_comp_prob = "{:.6f}".format(
                            float(expectedline_comp.split("::")[0])
                        )
                        # Update the expected component probability
                        expectedline_comp = (
                                rounded_expectedline_comp_prob
                                + "::"
                                + expectedline_comp.split("::")[1]
                        )
                        # If the learned probability is close enough to the expected probability
                        if (
                                abs(
                                    float(rounded_outline_comp_prob)
                                    - float(rounded_expectedline_comp_prob)
                                )
                                < 0.00001
                        ):
                            # Make the two lines identical
                            outline_comp = (
                                    rounded_expectedline_comp_prob
                                    + "::"
                                    + outline_comp.split("::")[1]
                            )
                    new_outline_comps.append(outline_comp)
                    new_expectedline_comps.append(expectedline_comp)
                new_outline = "; ".join(new_outline_comps)
                new_expectedline = "; ".join(new_expectedline_comps)
                expectedline = new_expectedline
                outline = new_outline
            self.assertEqual(outline, expectedline)

    return test


def main():
    AD_filenames = glob.glob(root_path("test", "lfi", "ad", "*.pl"))
    simple_filenames = glob.glob(root_path("test", "lfi", "simple", "*.pl"))
    misc_filenames = glob.glob(root_path("test", "lfi", "misc", "*.pl"))
    vars_in_T_filenames = glob.glob(root_path("test", "lfi", "vars_in_tunable", "*.pl"))

    evaluatables = ["ddnnf"]

    if has_sdd:
        evaluatables.append("sdd")
        evaluatables.append("sddx")

    else:
        print("No SDD support - The system tests are not performed with SDDs.")

    # tests for ADs
    for testfile in AD_filenames:
        testname = "test_lfi_ad_" + os.path.splitext(os.path.basename(testfile))[0]
        setattr(TestLFI, testname, createTestLFI(testfile, evaluatables))

    # tests for simple unit tests
    for testfile in simple_filenames:
        testname = "test_lfi_simple_" + os.path.splitext(os.path.basename(testfile))[0]
        setattr(TestLFI, testname, createTestLFI(testfile, evaluatables))

    # tests for Variables in t()
    for testfile in vars_in_T_filenames:
        testname = "test_lfi_vars_inT_" + os.path.splitext(os.path.basename(testfile))[0]
        setattr(TestLFI, testname, createTestLFI(testfile, evaluatables))

    # tests for Miscellaneous files
    for testfile in misc_filenames:
        testname = "test_lfi_misc_" + os.path.splitext(os.path.basename(testfile))[0]
        setattr(TestLFI, testname, createTestLFI(testfile, evaluatables))


main()

if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestLFI)
    unittest.TextTestRunner(verbosity=2).run(suite)



"""
problog.maxsat - Interface to MaxSAT solvers
--------------------------------------------

Interface to MaxSAT solvers.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from . import root_path
from .errors import ProbLogError
from .util import mktempfile, subprocess_check_output, Timer


class UnsatisfiableError(ProbLogError):
    def __init__(self):
        ProbLogError.__init__(
            self, "No solution exists that satisfies the constraints."
        )


class MaxSATSolver(object):
    def __init__(self, command):
        self.command = command

    @property
    def extension(self):
        return "cnf"

    def prepare_input(self, formula, **kwargs):
        return formula.to_dimacs(weighted=int, **kwargs)

    def process_output(self, output):
        for line in output.split("\n"):
            if line.startswith("v "):
                return list(map(int, line.split()[1:-1]))
        raise UnsatisfiableError()

    def call_process(self, inputf):
        filename = mktempfile("." + self.extension)
        with open(filename, "w") as f:
            f.write(inputf)
        return subprocess_check_output(self.command + [filename])

    def evaluate(self, formula, **kwargs):
        with Timer("Transform input"):
            inputf = self.prepare_input(formula, **kwargs)
        with Timer("Solver call"):
            output = self.call_process(inputf)
        with Timer("Transform output"):
            result = self.process_output(output)
        return result


class MIPMaxSATSolver(MaxSATSolver):
    def __init__(self, command):
        MaxSATSolver.__init__(self, command)

    @property
    def extension(self):
        return "lp"

    def prepare_input(self, formula, **kwargs):
        return formula.to_lp(**kwargs)


class SCIPSolver(MIPMaxSATSolver):
    def __init__(self):
        MaxSATSolver.__init__(self, ["scip", "-f"])

    def process_output(self, output):
        facts = set()
        in_the_zone = False
        for line in output.split("\n"):
            line = line.strip()
            if line.startswith("objective value"):
                in_the_zone = True
            elif in_the_zone:
                if not line:
                    return list(facts)
                else:
                    facts.add(int(line.split()[0][1:]))
        raise UnsatisfiableError()


def get_solver(prefer=None):
    if prefer == "scip":
        return SCIPSolver()
    elif prefer == "sat4j":
        return MaxSATSolver(
            ["java", "-jar", root_path("problog", "bin", "java", "sat4j-maxsat.jar")]
        )
    else:
        return MaxSATSolver(["maxsatz"])


def get_available_solvers():
    # TODO check whether they are actually available
    return ["maxsatz", "scip", "sat4j"]



import unittest

from problog import get_evaluatable
from problog.evaluator import SemiringProbability
from problog.formula import LogicFormula
from problog.program import PrologString

# noinspection PyBroadException
try:
    from pysdd import sdd

    has_sdd = True
except Exception as err:
    has_sdd = False

evaluatables = ["ddnnf"]

if has_sdd:
    evaluatables.append("sdd")
    evaluatables.append("sddx")
    evaluatables.append("fsdd")
else:
    print("No SDD support - The transformation tests are not performed with SDDs.")


class TestTransformation(unittest.TestCase):
    def setUp(self):
        try:
            self.assertCollectionEqual = self.assertItemsEqual
        except AttributeError:
            self.assertCollectionEqual = self.assertCountEqual

    def test_ad_atom_duplicate(self):
        """
        Run ad_atom_duplicate for each evaluatable.
        This test must pickup the case where during the transformation, additional _extra atoms are created because
        add_atom(..., cr_extra=True) is used instead of cr_extra=False.
        """
        for eval_name in evaluatables:
            with self.subTest(eval_name=eval_name):
                self.ad_atom_duplicate(eval_name)

    def ad_atom_duplicate(self, eval_name=None):
        """
        This test must pickup the case where during the transformation, additional _extra atoms are created because
        add_atom(..., cr_extra=True) is used instead of cr_extra=False.
        """
        program = """
                    0.2::a ; 0.8::b.
                    query(a).
                    query(b).
                """
        pl = PrologString(program)
        lf = LogicFormula.create_from(pl, label_all=True, avoid_name_clash=True)
        semiring = SemiringProbability()
        kc_class = get_evaluatable(name=eval_name, semiring=semiring)
        kc = kc_class.create_from(lf)  # type: LogicFormula
        self.assertEqual(3, kc.atomcount)


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestTransformation)
    unittest.TextTestRunner(verbosity=2).run(suite)



"""
ProbLog command-line interface.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import sys

from problog.cnf_formula import CNF
from problog.errors import process_error
from problog.evaluator import SemiringLogProbability
from problog.formula import LogicDAG, LogicFormula, LogicNNF
from problog.parser import DefaultPrologParser
from problog.program import ExtendedPrologFactory, PrologFile
from problog.util import subprocess_check_output, mktempfile


def get_arg_parser():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "filename", metavar="MODEL", type=str, help="input ProbLog model"
    )
    parser.add_argument(
        "--format",
        choices=("dot", "pl", "cnf", "svg", "internal"),
        default=None,
        help="output format",
    )
    parser.add_argument(
        "--break-cycles", action="store_true", help="perform cycle breaking"
    )
    parser.add_argument("--transform-nnf", action="store_true", help="transform to NNF")
    parser.add_argument(
        "--keep-all", action="store_true", help="also output deterministic nodes"
    )
    parser.add_argument(
        "--keep-duplicates",
        action="store_true",
        help="don't eliminate duplicate literals",
    )
    parser.add_argument(
        "--any-order", action="store_true", help="allow reordering nodes"
    )
    parser.add_argument(
        "--hide-builtins",
        action="store_true",
        help="hide deterministic part based on builtins",
    )
    parser.add_argument(
        "--propagate-evidence", action="store_true", help="propagate evidence"
    )
    parser.add_argument(
        "--propagate-weights", action="store_true", help="propagate evidence"
    )
    parser.add_argument(
        "--compact",
        action="store_true",
        help="allow compact model (may remove some predicates)",
    )
    parser.add_argument("--noninterpretable", action="store_true")
    parser.add_argument("--web", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument(
        "--verbose", "-v", action="count", default=0, help="Verbose output"
    )
    parser.add_argument("-o", "--output", type=str, help="output file", default=None)
    parser.add_argument(
        "-a",
        "--arg",
        dest="args",
        action="append",
        help="Pass additional arguments to the cmd_args builtin.",
    )
    return parser


def main(argv, result_handler=None):

    args = get_arg_parser().parse_args(argv)

    outformat = args.format
    outfile = sys.stdout
    if args.output:
        outfile = open(args.output, "w")
        if outformat is None:
            outformat = os.path.splitext(args.output)[1][1:]

    if outformat == "cnf" and not args.break_cycles:
        print(
            "Warning: CNF output requires cycle-breaking; cycle breaking enabled.",
            file=sys.stderr,
        )

    if args.transform_nnf:
        target = LogicNNF
    elif args.break_cycles or outformat == "cnf":
        target = LogicDAG
    else:
        target = LogicFormula

    if args.propagate_weights:
        semiring = SemiringLogProbability()
    else:
        semiring = None

    if args.web:
        print_result = print_result_json
    else:
        print_result = print_result_standard

    final_result = None
    try:
        gp = target.createFrom(
            PrologFile(
                args.filename, parser=DefaultPrologParser(ExtendedPrologFactory())
            ),
            label_all=not args.noninterpretable,
            avoid_name_clash=not args.compact,
            keep_order=not args.any_order,
            keep_all=args.keep_all,
            keep_duplicates=args.keep_duplicates,
            hide_builtins=args.hide_builtins,
            propagate_evidence=args.propagate_evidence,
            propagate_weights=semiring,
            args=args.args,
        )

        if outformat == "pl":
            final_result = (True, gp.to_prolog())
        elif outformat == "dot":
            final_result = (True, gp.to_dot())
        elif outformat == "svg":
            dot = gp.to_dot()
            tmpfile = mktempfile(".dot")
            with open(tmpfile, "w") as f:
                print(dot, file=f)
            svg = subprocess_check_output(["dot", tmpfile, "-Tsvg"])
            final_result = (True, gp.to_dot())
        elif outformat == "cnf":
            cnfnames = False
            if args.verbose > 0:
                cnfnames = True
            final_result = (True, CNF.createFrom(gp).to_dimacs(names=cnfnames))
        elif outformat == "internal":
            final_result = (True, str(gp))
        else:
            final_result = (True, gp.to_prolog())
    except Exception as err:
        import traceback

        err.trace = traceback.format_exc()
        final_result = (False, err)

    rc = print_result(final_result, output=outfile)
    if args.output:
        outfile.close()

    if rc:
        sys.exit(rc)

    return final_result


def print_result_standard(result, output=sys.stdout):
    success, result = result
    if success:
        print(result, file=output)
        return 0
    else:
        print(process_error(result), file=output)
        return 1


def print_result_json(result, output=sys.stdout):
    success, result = result
    import json

    out = {"SUCCESS": success}
    if success:
        out["result"] = str(result)
    else:
        out["err"] = vars(result)
    print(json.dumps(out), file=output)
    return 0



"""
Part of the ProbLog distribution.

Copyright 2022 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import unittest

from problog import get_evaluatable
from problog.program import PrologString


class TestLibString(unittest.TestCase):
    def test_str2lst(self):
        p = PrologString(
            """
                    :- use_module(library(string)).
                    test(L) :- str2lst('aaa zzz eee rrr', L).
                    query(test(L)).
                """
        )

        res = get_evaluatable().create_from(p).evaluate()
        self.assertEqual(list(res.values()), [1.0])  # probability check
        self.assertEqual(
            [str(x) for x in res.keys()],
            ["test([''', a, a, a, ' ', z, z, z, ' ', e, e, e, ' ', r, r, " "r, '''])"],
        )  # name check

    def test_lst2str(self):
        p = PrologString(
            """
                    :- use_module(library(string)).
                    test(L) :- lst2str(['aaa','dddd','fffff'], L).
                    query(test(L)).
                """
        )

        res = get_evaluatable().create_from(p).evaluate()
        self.assertEqual(list(res.values()), [1.0])  # probability check
        self.assertEqual(
            [str(x) for x in res.keys()], ["test(aaaddddfffff)"]
        )  # name check

    def test_join(self):
        p = PrologString(
            """
                    :- use_module(library(string)).
                    test(L) :- join(',', ['aaa','dddd','fffff'], L).
                    query(test(L)).
                """
        )

        res = get_evaluatable().create_from(p).evaluate()
        self.assertEqual(list(res.values()), [1])  # probability check
        self.assertEqual(
            [str(x) for x in res.keys()], ["test('aaa,dddd,fffff')"]
        )  # name check

    def test_concat(self):
        p = PrologString(
            """
                    :- use_module(library(string)).
                    test(L) :- concat(['aaa','dddd','fffff'], L).
                    query(test(L)).
                """
        )
        res = get_evaluatable().create_from(p).evaluate()
        self.assertEqual(list(res.values()), [1.0])  # probability check
        self.assertEqual(
            [str(x) for x in res.keys()], ["test(aaaddddfffff)"]
        )  # name check

    def test_numerical_string(self):
        p = PrologString(
            '''
                    :- use_module(library(string)).
                    query(str2int("2",X)).
            '''
        )
        res = get_evaluatable().create_from(p).evaluate()
        self.assertEqual(list(res.values()), [1.0])  # probability check
        self.assertEqual(
            [str(x) for x in res.keys()], ['str2int("2",2)']
        )


import glob
import sys
import unittest
from copy import deepcopy

from problog import root_path
from problog import tasks


def read_result(filename):
    results = []
    with open(filename) as f:
        result = {}
        reading = False
        for l in f:
            l = l.strip()
            if l.startswith("%Expected outcome:"):
                reading = True
            elif reading:
                if l.lower().startswith("% error"):
                    return l[len("% error") :].strip()
                elif l.startswith("% "):
                    try:
                        query, prob = l[2:].rsplit(None, 1)
                        result[query.strip()] = float(prob.strip())
                    except:
                        results.append(deepcopy(result))
                        result = {}
                else:
                    reading = False
            if l.startswith("query(") and l.find("% outcome:") >= 0:
                pos = l.find("% outcome:")
                query = l[6:pos].strip().rstrip(".").rstrip()[:-1]
                prob = l[pos + 10 :]
                result[query.strip()] = float(prob.strip())
        results.append(result)
    return results


class TestConstraints(unittest.TestCase):
    def test_constraints(self):
        for filename in glob.glob(root_path("test/constraints", "*.pl")):
            try:
                solutions = tasks.load_task("constraint").run(filename)
            except ImportError:
                sys.stderr.write(
                    "No flatzinc support - The constraint tests are not performed.\n"
                )
                return True

            solutions = [
                {str(k): v for k, v in solution if v > 0.0} for solution in solutions
            ]
            expected = read_result(filename)
            self.assertTrue(solutions == expected)



import os
import unittest
from pathlib import Path

from problog.logic import Constant, Term, Not
from problog.tasks import map, explain, time1, bayesnet, mpe, ground, probability

dirname = os.path.dirname(__file__)
test_folder = Path(dirname, "./../../test/")


class TestTasks(unittest.TestCase):
    def test_map(self):
        file_name = test_folder / "tasks" / "map_probabilistic_graph.pl"
        result = map.main([str(file_name)])
        success = result[0]
        if not success:
            self.fail("Failed executing MAP on " + str(file_name))
        else:
            choices, score, stats = result[1]
            self.assertEqual(
                {
                    Term("edge", Constant(1), Constant(2)): 1,
                    Term("edge", Constant(1), Constant(3)): 0,
                },
                choices,
            )
            self.assertAlmostEqual(1.878144, score, places=5)

    # def test_mpe_semiring_some_heads(self):
    #     file_name = test_folder / "tasks" / "some_heads.pl"
    #     result = mpe.main([str(file_name), "--use-semiring"])
    #     self.assertTrue(result[0])
    #     self.assertEqual(0.3, result[1][0])
    #     self.assertEqual([Term("someHeads")], result[1][1])

    def test_mpe_maxsat_some_heads(self):
        file_name = test_folder / "tasks" / "some_heads.pl"
        result = mpe.main([str(file_name), "--use-maxsat"])
        self.assertTrue(result[0])
        self.assertAlmostEqual(0.3, result[1][0], places=6)
        self.assertEqual([Term("someHeads")], result[1][1])

    def test_mpe_maxsat_pgraph(self):
        file_name = test_folder / "tasks" / "pgraph.pl"
        result = mpe.main([str(file_name), "--use-maxsat"])
        self.assertTrue(result[0])
        self.assertAlmostEqual(0.3, result[1][0], delta=1e6)
        self.assertEqual(
            {
                Term("edge", Constant(1), Constant(2)),
                Term("edge", Constant(2), Constant(5)),
                Not("\\+", Term("edge", Constant(1), Constant(3))),
                Not("\\+", Term("edge", Constant(3), Constant(4))),
                Term("edge", Constant(4), Constant(5)),
                Term("edge", Constant(2), Constant(6)),
                Not("\\+", Term("edge", Constant(5), Constant(6))),
            },
            set(result[1][1]),
        )

    def test_explain_trivial(self):
        file_name = test_folder / "tasks" / "map_probabilistic_graph.pl"
        result = explain.main([str(file_name)])

        self.assertTrue(result["SUCCESS"])
        results = result["results"]
        self.assertAlmostEqual(
            0.6, results[Term("edge", Constant(1), Constant(2))], delta=1e6
        )
        self.assertAlmostEqual(
            0.1, results[Term("edge", Constant(1), Constant(3))], delta=1e6
        )

    def test_explain_some_heads(self):
        file_name = test_folder / "tasks" / "some_heads.pl"
        result = explain.main([str(file_name)])

        # Check if successful
        self.assertTrue(result["SUCCESS"])

        # Test proofs
        proofs = result["proofs"]
        self.assertEqual("someHeads :- heads2.  % P=0.6", proofs[0])
        self.assertEqual("someHeads :- heads1, \\+heads2.  % P=0.2", proofs[1])

        # Test result
        results = result["results"]
        self.assertAlmostEqual(0.8, results[Term("someHeads")], delta=1e6)

    def check_probability(self, expected, result):
        self.assertTrue(result[0])
        self.assertEqual(expected, result[1])

    def test_probability_some_heads(self):
        file_name = test_folder / "tasks" / "some_heads.pl"
        result = probability.main_result([str(file_name)])
        self.check_probability({Term("someHeads"): 0.8}, result)

    def check_probability_probabilistic_graph(self, result):
        self.assertTrue(result[0])
        self.assertAlmostEqual(
            0.11299610205528002,
            result[1][Term("edge", Constant(1), Constant(3))],
            delta=1e6,
        )
        self.assertAlmostEqual(
            0.9911410347271441,
            result[1][Term("edge", Constant(1), Constant(2))],
            delta=1e6,
        )

    def test_probability_pgraph(self):
        file_name = test_folder / "tasks" / "map_probabilistic_graph.pl"
        self.check_probability_probabilistic_graph(probability.main_result([str(file_name)]))
        self.check_probability_probabilistic_graph(
            probability.main_result([str(file_name), "--combine"])
        )
        self.check_probability_probabilistic_graph(
            probability.main_result([str(file_name), "--nologspace"])
        )
        self.check_probability_probabilistic_graph(
            probability.main_result([str(file_name), "--propagate-evidence"])
        )
        self.check_probability_probabilistic_graph(
            probability.main_result([str(file_name), "--propagate-weights"])
        )
        self.check_probability_probabilistic_graph(
            probability.main_result(
                [str(file_name), "--propagate-evidence", "--propagate-weights"]
            )
        )
        self.check_probability_probabilistic_graph(
            probability.main_result([str(file_name), "--unbuffered"])
        )
        self.check_probability_probabilistic_graph(
            probability.main_result([str(file_name), "--convergence", str(0.00000001)])
        )
        self.check_probability_probabilistic_graph(
            probability.main_result([str(file_name), "--format", "prolog"])
        )
        self.check_probability_probabilistic_graph(
            probability.main_result([str(file_name), "--web"])
        )

    def check_ground_result(self, expected, result):
        self.assertTrue(result[0])
        print("result", result)
        self.assertEqual(expected, result[1])

    def test_ground(self):
        file_name = test_folder / "tasks" / "pgraph.pl"
        grounded_pgraph = (
            "0.6::edge(1,2)."
            "\n0.1::edge(1,3)."
            "\n0.4::edge(2,5)."
            "\n0.3::edge(2,6)."
            "\npath(2,5) :- edge(2,5)."
            "\n0.3::edge(3,4)."
            "\n0.8::edge(4,5)."
            "\npath(4,5) :- edge(4,5)."
            "\npath(3,5) :- edge(3,4), path(4,5)."
            "\npath(1,5) :- edge(1,2), path(2,5)."
            "\npath(1,5) :- edge(1,3), path(3,5)."
            "\n0.2::edge(5,6)."
            "\npath(5,6) :- edge(5,6)."
            "\npath(2,6) :- edge(2,6)."
            "\npath(2,6) :- edge(2,5), path(5,6)."
            "\npath(4,6) :- edge(4,5), path(5,6)."
            "\npath(3,6) :- edge(3,4), path(4,6)."
            "\npath(1,6) :- edge(1,2), path(2,6)."
            "\npath(1,6) :- edge(1,3), path(3,6)."
            "\nevidence(path(1,5))."
            "\nevidence(path(1,6))."
        )
        self.check_ground_result(grounded_pgraph, ground.main([str(file_name)]))
        self.check_ground_result(
            grounded_pgraph, ground.main([str(file_name), "--propagate-evidence"])
        )
        self.check_ground_result(
            grounded_pgraph, ground.main([str(file_name), "--propagate-weights"])
        )
        self.check_ground_result(
            grounded_pgraph, ground.main([str(file_name), "--hide-builtins"])
        )
        self.check_ground_result(
            "0.6::edge(1,2).\n0.1::edge(1,3).\n0.4::edge(2,5).\n0.3::edge(2,6).\n0.3::edge(3,4).\n0.8::edge(4,"
            "5).\npath(3,5) :- edge(3,4), edge(4,5).\npath(1,5) :- edge(1,2), edge(2,5).\npath(1,5) :- edge(1,3), "
            "path(3,5).\n0.2::edge(5,6).\npath(2,6) :- edge(2,6).\npath(2,6) :- edge(2,5), edge(5,6).\npath(4,"
            "6) :- edge(4,5), edge(5,6).\npath(3,6) :- edge(3,4), path(4,6).\npath(1,6) :- edge(1,2), path(2,"
            "6).\npath(1,6) :- edge(1,3), path(3,6).\nevidence(path(1,5)).\nevidence(path(1,6)).",
            ground.main([str(file_name), "--compact"]),
        )
        self.check_ground_result(
            "0.6::edge(1,2).\n0.1::edge(1,3).\n0.4::edge(2,5).\n0.3::edge(2,6).\n0.3::edge(3,4).\n0.8::edge(4,"
            "5).\npath(1,5) :- edge(1,2), edge(2,5).\npath(1,5) :- edge(1,3), edge(3,4), edge(4,5).\n0.2::edge(5,"
            "6).\nNone :- edge(2,6).\nNone :- edge(2,5), edge(5,6).\npath(1,6) :- edge(1,2), edge(2,6).\npath(1,"
            "6) :- edge(1,3), edge(3,4), edge(4,5), edge(5,6).\nevidence(path(1,5)).\nevidence(path(1,6)).",
            ground.main([str(file_name), "--noninterpretable"]),
        )

    def test_ground_formats_no_error(self):
        file_name = test_folder / "tasks" / "pgraph.pl"

        result = ground.main([str(file_name), "--format", "pl"])
        # result = ground.main([str(file_name), "--format", "dot"]) # Broken
        # result = ground.main([str(file_name), "--format", "svg"])
        result = ground.main([str(file_name), "--format", "cnf"])
        result = ground.main([str(file_name), "--format", "internal"])
        result = ground.main([str(file_name), "--web"])

    def test_time(self):
        file_name = test_folder / "tasks" / "some_heads.pl"
        n = 3
        result = time1.main([str(file_name), f"-n {n}"])
        # One result for one file
        self.assertEqual(1, len(result))
        result = result[str(file_name)]
        # n results for n repetitions
        self.assertEqual(n, len(result))
        for run_n in result:
            # Check timers in each repetition
            expected_names = {"parse", "load", "ground", "cycles", "compile", "evaluate"}
            timer_names = {timer.name for timer in run_n}
            assert expected_names.issubset(timer_names)

    # BN
    def check_bn(self, file_name, expected):
        result = bayesnet.main([str(file_name)])
        self.assertTrue(result[0])
        self.equal_bn_result(expected, result[1])

    @staticmethod
    def normalise_bn_output(bn_string):
        return bn_string.strip().replace("\n\n", "\n").replace(" ", "")

    def equal_bn_result(self, expected, actual):
        return self.assertEqual(
            self.normalise_bn_output(expected), self.normalise_bn_output(actual)
        )

    def test_bn(self):
        file_name = test_folder / "bn" / "single_atom_ad.pl"
        self.check_bn(
            file_name,
            "Factor (c0) =0, 1"
            "\n(): [0.4, 0.6]"
            "\n\nOrCPT someHeads [0,1] -- c0"
            "\n('c0', 1)",
        )

    def test_bn_some_heads(self):
        file_name = test_folder / "tasks" / "some_heads.pl"
        self.check_bn(
            file_name,
            "Factor (c0) = 0, 1\n(): [0.5, 0.5]"
            "\n\nOrCPT heads1 [0,1] -- c0"
            "\n('c0', 1)"
            "\n\nFactor (c1) = 0, 1"
            "\n(): [0.4, 0.6]"
            "\n\nOrCPT heads2 [0,1] -- c1"
            "\n('c1', 1)"
            "\n\nFactor (c2 | heads1) = 0, 1"
            "\n(False,): [1.0, 0.0]\n(True,): [0.0, 1.0]"
            "\n\nOrCPT someHeads [0,1] -- c2,c3\n('c2', 1)\n('c3', 1)"
            "\n\nFactor (c3 | heads2) = 0, 1"
            "\n(False,): [1.0, 0.0]\n(True,): [0.0, 1.0]\n",
        )

    def test_bn_some_heads_or(self):
        file_name = test_folder / "tasks" / "some_heads_or.pl"
        self.check_bn(
            file_name,
            "Factor (c0) = 0, 1\n(): [0.5, 0.5]"
            "\n\nOrCPT heads1 [0,1] -- c0\n('c0', 1)"
            "\n\nFactor (c1) = 0, 1\n(): [0.4, 0.6]"
            "\n\nOrCPT heads2 [0,1] -- c1\n('c1', 1)"
            "\n\nFactor (c2 | heads2) = 0\n(False,): [1.0]\n(True,): [1.0]"
            "\n\nFactor (c3 | \\+heads1, \\+heads2) = 0\n(False,): [1.0]\n(True,): [1.0]"
            "\n\nFactor (c4 | heads1) = 0\n(False,): [1.0]\n(True,): [1.0]"
            "\n\nFactor (c5 | heads2) = 0\n(False,): [1.0]\n(True,): [1.0]"
            "\n\nFactor (c6 | heads1) = 0\n(False,): [1.0]\n(True,): [1.0]"
            "\n\nFactor (c7 | heads1, heads2) = 0, 1\n(False, False): [1.0, 0.0]"
            "\n(False, True): [1.0, 0.0]\n(True, False): [1.0, 0.0]\n(True, True): [0.0, 1.0]"
            "\n\nOrCPT someHeads [0,1] -- c7\n('c7', 1)\n",
        )

    def test_bn_pgraph(self):
        file_name = test_folder / "tasks" / "pgraph.pl"
        self.check_bn(
            file_name,
            "Factor (c0) = 0, 1"
            "\n(): [0.4, 0.6]"
            "\n\nOrCPT edge(1,2) [0,1] -- c0"
            "\n('c0', 1)"
            "\n\nFactor (c1) = 0, 1"
            "\n(): [0.6, 0.4]"
            "\n\nOrCPT edge(2,5) [0,1] -- c1"
            "\n('c1', 1)"
            "\n\nFactor (c2) = 0, 1"
            "\n(): [0.9, 0.1]"
            "\n\nOrCPT edge(1,3) [0,1] -- c2"
            "\n('c2', 1)\n\nFactor (c3) = 0, 1"
            "\n(): [0.7, 0.3]"
            "\n\nOrCPT edge(3,4) [0,1] -- c3\n('c3', 1)"
            "\n\nFactor (c4) = 0, 1\n(): [0.19999999999999996, 0.8]"
            "\n\nOrCPT edge(4,5) [0,1] -- c4\n('c4', 1)"
            "\n\nFactor (c5 | edge(3,4), edge(4,5)) = 0, 1\n(False, False): [1.0, 0.0]"
            "\n(False, True): [1.0, 0.0]\n(True, False): [1.0, 0.0]\n(True, True): [0.0, 1.0]"
            "\n\nOrCPT path(3,5) [0,1] -- c5\n('c5', 1)\n\nFactor (c6 | edge(1,2), edge(2,5)) = 0, 1"
            "\n(False, False): [1.0, 0.0]\n(False, True): [1.0, 0.0]\n(True, False): [1.0, 0.0]"
            "\n(True, True): [0.0, 1.0]\n\nOrCPT path(1,5) [0,1] -- c6,c7\n('c6', 1)\n('c7', 1)"
            "\n\nFactor (c7 | edge(1,3), path(3,5)) = 0, 1\n(False, False): [1.0, 0.0]\n(False, True): [1.0, 0.0]"
            "\n(True, False): [1.0, 0.0]\n(True, True): [0.0, 1.0]\n\nFactor (c8) = 0, 1\n(): [0.7, 0.3]"
            "\n\nOrCPT edge(2,6) [0,1] -- c8\n('c8', 1)\n\nFactor (c9) = 0, 1\n(): [0.8, 0.2]"
            "\n\nOrCPT edge(5,6) [0,1] -- c9\n('c9', 1)"
            "\n\nFactor (c10 | edge(2,6)) = 0, 1\n(False,): [1.0, 0.0]\n(True,): [0.0, 1.0]"
            "\n\nOrCPT path(2,6) [0,1] -- c10,c11\n('c10', 1)\n('c11', 1)"
            "\n\nFactor (c11 | edge(2,5), edge(5,6)) = 0, 1\n(False, False): [1.0, 0.0]"
            "\n(False, True): [1.0, 0.0]\n(True, False): [1.0, 0.0]\n(True, True): [0.0, 1.0]"
            "\n\nFactor (c12 | edge(4,5), edge(5,6)) = 0, 1"
            "\n(False, False): [1.0, 0.0]\n(False, True): [1.0, 0.0]"
            "\n(True, False): [1.0, 0.0]\n(True, True): [0.0, 1.0]"
            "\n\nOrCPT path(4,6) [0,1] -- c12\n('c12', 1)"
            "\n\nFactor (c13 | edge(3,4), path(4,6)) = 0, 1\n(False, False): [1.0, 0.0]"
            "\n(False, True): [1.0, 0.0]\n(True, False): [1.0, 0.0]\n(True, True): [0.0, 1.0]"
            "\n\nOrCPT path(3,6) [0,1] -- c13\n('c13', 1)"
            "\n\nFactor (c14 | edge(1,2), path(2,6)) = 0, 1"
            "\n(False, False): [1.0, 0.0]\n(False, True): [1.0, 0.0]"
            "\n(True, False): [1.0, 0.0]\n(True, True): [0.0, 1.0]"
            "\n\nOrCPT path(1,6) [0,1] -- c14,c15\n('c14', 1)\n('c15', 1)"
            "\n\nFactor (c15 | edge(1,3), path(3,6)) = 0, 1\n(False, False): [1.0, 0.0]"
            "\n(False, True): [1.0, 0.0]\n(True, False): [1.0, 0.0]\n(True, True): [0.0, 1.0]\n",
        )
